"""Methods to collect, analyze and manipulate transition and trajectory rollouts."""

import collections
import dataclasses
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from gymnasium import spaces
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.utils import check_for_correct_spaces
from stable_baselines3.common.vec_env import VecEnv

from imitation.data import types


def unwrap_traj(traj: types.TrajectoryWithRew) -> types.TrajectoryWithRew:
    """Uses `RolloutInfoWrapper`-captured `obs` and `rews` to replace fields.

    This can be useful for bypassing other wrappers to retrieve the original
    `obs` and `rews`.

    Fails if `infos` is None or if the trajectory was generated from an
    environment without imitation.data.wrappers.RolloutInfoWrapper

    Args:
        traj: A trajectory generated from `RolloutInfoWrapper`-wrapped Environments.

    Returns:
        A copy of `traj` with replaced `obs` and `rews` fields.

    Raises:
        ValueError: If `traj.infos` is None
    """
    if traj.infos is None:
        raise ValueError("Trajectory must have infos to unwrap")
    ep_info = traj.infos[-1]["rollout"]
    res = dataclasses.replace(traj, obs=ep_info["obs"], rews=ep_info["rews"])
    assert len(res.obs) == len(res.acts) + 1
    assert len(res.rews) == len(res.acts)
    return res


class TrajectoryAccumulator:
    """Accumulates trajectories step-by-step.

    Useful for collecting completed trajectories while ignoring partially-completed
    trajectories (e.g. when rolling out a VecEnv to collect a set number of
    transitions). Each in-progress trajectory is identified by a 'key', which enables
    several independent trajectories to be collected at once. They key can also be left
    at its default value of `None` if you only wish to collect one trajectory.
    """

    def __init__(self):
        """Initialise the trajectory accumulator."""
        self.partial_trajectories = collections.defaultdict(list)

    def add_step(
        self,
        step_dict: Mapping[str, Union[types.Observation, Mapping[str, Any]]],
        key: Hashable = None,
    ) -> None:
        """Add a single step to the partial trajectory identified by `key`.

        Generally a single step could correspond to, e.g., one environment managed
        by a VecEnv.

        Args:
            step_dict: dictionary containing information for the current step. Its
                keys could include any (or all) attributes of a `TrajectoryWithRew`
                (e.g. "obs", "acts", etc.).
            key: key to uniquely identify the trajectory to append to, if working
                with multiple partial trajectories.
        """
        self.partial_trajectories[key].append(step_dict)

    def finish_trajectory(
        self,
        key: Hashable,
        terminal: bool,
    ) -> types.TrajectoryWithRew:
        """Complete the trajectory labelled with `key`.

        Args:
            key: key uniquely identifying which in-progress trajectory to remove.
            terminal: trajectory has naturally finished (i.e. includes terminal state).

        Returns:
            traj: list of completed trajectories popped from
                `self.partial_trajectories`.
        """
        part_dicts = self.partial_trajectories[key]
        del self.partial_trajectories[key]
        out_dict_unstacked = collections.defaultdict(list)
        for part_dict in part_dicts:
            for k, array in part_dict.items():
                out_dict_unstacked[k].append(array)

        out_dict_stacked = {
            k: types.stack_maybe_dictobs(arr_list)
            for k, arr_list in out_dict_unstacked.items()
        }
        traj = types.TrajectoryWithRew(**out_dict_stacked, terminal=terminal)
        assert traj.rews.shape[0] == traj.acts.shape[0] == len(traj.obs) - 1
        return traj

    def add_steps_and_auto_finish(
        self,
        acts: np.ndarray,
        obs: Union[types.Observation, Dict[str, np.ndarray]],
        rews: np.ndarray,
        dones: np.ndarray,
        infos: List[dict],
    ) -> List[types.TrajectoryWithRew]:
        """Calls `add_step` repeatedly using acts and the returns from `venv.step`.

        Also automatically calls `finish_trajectory()` for each `done == True`.
        Before calling this method, each environment index key needs to be
        initialized with the initial observation (usually from `venv.reset()`).

        See the body of `util.rollout.generate_trajectory` for an example.

        Args:
            acts: Actions passed into `VecEnv.step()`.
            obs: Return value from `VecEnv.step(acts)`.
            rews: Return value from `VecEnv.step(acts)`.
            dones: Return value from `VecEnv.step(acts)`.
            infos: Return value from `VecEnv.step(acts)`.

        Returns:
            A list of completed trajectories. There should be one trajectory for
            each `True` in the `dones` argument.
        """
        trajs: List[types.TrajectoryWithRew] = []
        wrapped_obs = types.maybe_wrap_in_dictobs(obs)

        # iterate through environments
        for env_idx in range(len(wrapped_obs)):
            assert env_idx in self.partial_trajectories
            assert list(self.partial_trajectories[env_idx][0].keys()) == ["obs"], (
                "Need to first initialize partial trajectory using "
                "self._traj_accum.add_step({'obs': ob}, key=env_idx)"
            )

        # iterate through steps
        zip_iter = enumerate(zip(acts, wrapped_obs, rews, dones, infos))
        for env_idx, (act, ob, rew, done, info) in zip_iter:
            if done:
                # When dones[i] from VecEnv.step() is True, obs[i] is the first
                # observation following reset() of the ith VecEnv, and
                # infos[i]["terminal_observation"] is the actual final observation.
                real_ob = types.maybe_wrap_in_dictobs(info["terminal_observation"])
            else:
                real_ob = ob

            self.add_step(
                dict(
                    acts=act,
                    rews=rew,
                    # this is not the obs corresponding to `act`, but rather the obs
                    # *after* `act` (see above)
                    obs=real_ob,
                    infos=info,
                ),
                env_idx,
            )
            if done:
                # finish env_idx-th trajectory
                new_traj = self.finish_trajectory(env_idx, terminal=True)
                trajs.append(new_traj)
                # When done[i] from VecEnv.step() is True, obs[i] is the first
                # observation following reset() of the ith VecEnv.
                self.add_step(dict(obs=ob), env_idx)
        return trajs


GenTrajTerminationFn = Callable[[Sequence[types.TrajectoryWithRew]], bool]


def make_min_episodes(n: int) -> GenTrajTerminationFn:
    """Terminate after collecting n episodes of data.

    Args:
        n: Minimum number of episodes of data to collect.
            May overshoot if two episodes complete simultaneously (unlikely).

    Returns:
        A function implementing this termination condition.
    """
    assert n >= 1
    return lambda trajectories: len(trajectories) >= n


def make_min_timesteps(n: int) -> GenTrajTerminationFn:
    """Terminate at the first episode after collecting n timesteps of data.

    Args:
        n: Minimum number of timesteps of data to collect.
            May overshoot to nearest episode boundary.

    Returns:
        A function implementing this termination condition.
    """
    assert n >= 1

    def f(trajectories: Sequence[types.TrajectoryWithRew]):
        timesteps = sum(len(t.obs) - 1 for t in trajectories)
        return timesteps >= n

    return f


def make_sample_until(
    min_timesteps: Optional[int] = None,
    min_episodes: Optional[int] = None,
) -> GenTrajTerminationFn:
    """Returns a termination condition sampling for a number of timesteps and episodes.

    Args:
        min_timesteps: Sampling will not stop until there are at least this many
            timesteps.
        min_episodes: Sampling will not stop until there are at least this many
            episodes.

    Returns:
        A termination condition.

    Raises:
        ValueError: Neither of n_timesteps and n_episodes are set, or either are
            non-positive.
    """
    if min_timesteps is None and min_episodes is None:
        raise ValueError(
            "At least one of min_timesteps and min_episodes needs to be non-None",
        )

    conditions = []
    if min_timesteps is not None:
        if min_timesteps <= 0:
            raise ValueError(
                f"min_timesteps={min_timesteps} if provided must be positive",
            )
        conditions.append(make_min_timesteps(min_timesteps))

    if min_episodes is not None:
        if min_episodes <= 0:
            raise ValueError(
                f"min_episodes={min_episodes} if provided must be positive",
            )
        conditions.append(make_min_episodes(min_episodes))

    def sample_until(trajs: Sequence[types.TrajectoryWithRew]) -> bool:
        for cond in conditions:
            if not cond(trajs):
                return False
        return True

    return sample_until


# A PolicyCallable is a function that takes an array of observations, an optional
# array of states, and an optional array of episode starts and returns an array of
# corresponding actions.
PolicyCallable = Callable[
    [
        Union[np.ndarray, Dict[str, np.ndarray]],  # observations
        Optional[Tuple[np.ndarray, ...]],  # states
        Optional[np.ndarray],  # episode_starts
    ],
    Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]],  # actions, states
]
AnyPolicy = Union[BaseAlgorithm, BasePolicy, PolicyCallable, None]


def policy_to_callable(
    policy: AnyPolicy,
    venv: VecEnv,
    deterministic_policy: bool = False,
) -> PolicyCallable:
    """Converts any policy-like object into a function from observations to actions."""
    get_actions: PolicyCallable
    if policy is None:

        def get_actions(
            observations: Union[np.ndarray, Dict[str, np.ndarray]],
            states: Optional[Tuple[np.ndarray, ...]],
            episode_starts: Optional[np.ndarray],
        ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
            acts = [venv.action_space.sample() for _ in range(len(observations))]
            return np.stack(acts, axis=0), None

    elif isinstance(policy, (BaseAlgorithm, BasePolicy)):
        # There's an important subtlety here: BaseAlgorithm and BasePolicy
        # are themselves Callable (which we check next). But in their case,
        # we want to use the .predict() method, rather than __call__()
        # (which would call .forward()). So this elif clause must come first!

        def get_actions(
            observations: Union[np.ndarray, Dict[str, np.ndarray]],
            states: Optional[Tuple[np.ndarray, ...]],
            episode_starts: Optional[np.ndarray],
        ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
            assert isinstance(policy, (BaseAlgorithm, BasePolicy))
            # pytype doesn't seem to understand that policy is a BaseAlgorithm
            # or BasePolicy here, rather than a Callable
            (acts, states) = policy.predict(  # pytype: disable=attribute-error
                observations,
                state=states,
                episode_start=episode_starts,
                deterministic=deterministic_policy,
            )
            return acts, states

    elif callable(policy):
        # When a policy callable is passed, by default we will use it directly.
        # We are not able to change the determinism of the policy when it is a
        # callable that only takes in the states.
        if deterministic_policy:
            raise ValueError(
                "Cannot set deterministic_policy=True when policy is a callable, "
                "since deterministic_policy argument is ignored.",
            )
        get_actions = policy

    else:
        raise TypeError(
            "Policy must be None, a stable-baselines policy or algorithm, "
            f"or a Callable, got {type(policy)} instead",
        )

    if isinstance(policy, BaseAlgorithm):
        # Check that the observation and action spaces of policy and environment match
        try:
            check_for_correct_spaces(
                venv,
                policy.observation_space,
                policy.action_space,
            )
        except ValueError as e:
            # Check for a particularly common mistake when using image environments.
            venv_obs_shape = venv.observation_space.shape
            assert policy.observation_space is not None
            policy_obs_shape = policy.observation_space.shape
            assert venv_obs_shape is not None
            assert policy_obs_shape is not None
            if len(venv_obs_shape) != 3 or len(policy_obs_shape) != 3:
                raise e
            venv_obs_rearranged = (
                venv_obs_shape[2],
                venv_obs_shape[0],
                venv_obs_shape[1],
            )
            if venv_obs_rearranged != policy_obs_shape:
                raise e
            raise ValueError(
                "Policy and environment observation shape mismatch. "
                "This is likely caused by "
                "https://github.com/HumanCompatibleAI/imitation/issues/599. "
                "If encountering this from rollout.rollout, try calling:\n"
                "rollout.rollout(expert, expert.get_env(), ...) instead of\n"
                "rollout.rollout(expert, env, ...)\n\n"
                f"Policy observation shape: {policy_obs_shape} \n"
                f"Environment observation shape: {venv_obs_shape}",
            )

    return get_actions


def generate_trajectories(
    policy: AnyPolicy,
    venv: VecEnv,
    sample_until: GenTrajTerminationFn,
    rng: np.random.Generator,
    *,
    deterministic_policy: bool = False,
) -> Sequence[types.TrajectoryWithRew]:
    """Generate trajectory dictionaries from a policy and an environment.

    Args:
        policy: Can be any of the following:
            1) A stable_baselines3 policy or algorithm trained on the gym environment.
            2) A Callable that takes an ndarray of observations and returns an ndarray
            of corresponding actions.
            3) None, in which case actions will be sampled randomly.
        venv: The vectorized environments to interact with.
        sample_until: A function determining the termination condition.
            It takes a sequence of trajectories, and returns a bool.
            Most users will want to use one of `min_episodes` or `min_timesteps`.
        deterministic_policy: If True, asks policy to deterministically return
            action. Note the trajectories might still be non-deterministic if the
            environment has non-determinism!
        rng: used for shuffling trajectories.

    Returns:
        Sequence of trajectories, satisfying `sample_until`. Additional trajectories
        may be collected to avoid biasing process towards short episodes; the user
        should truncate if required.
    """
    get_actions = policy_to_callable(policy, venv, deterministic_policy)

    # Collect rollout tuples.
    trajectories = []
    # accumulator for incomplete trajectories
    trajectories_accum = TrajectoryAccumulator()
    obs = venv.reset()
    assert isinstance(
        obs,
        (np.ndarray, dict),
    ), "Tuple observations are not supported."
    wrapped_obs = types.maybe_wrap_in_dictobs(obs)

    # we use dictobs to iterate over the envs in a vecenv
    for env_idx, ob in enumerate(wrapped_obs):
        # Seed with first obs only. Inside loop, we'll only add second obs from
        # each (s,a,r,s') tuple, under the same "obs" key again. That way we still
        # get all observations, but they're not duplicated into "next obs" and
        # "previous obs" (this matters for, e.g., Atari, where observations are
        # really big).
        trajectories_accum.add_step(dict(obs=ob), env_idx)

    # Now, we sample until `sample_until(trajectories)` is true.
    # If we just stopped then this would introduce a bias towards shorter episodes,
    # since longer episodes are more likely to still be active, i.e. in the process
    # of being sampled from. To avoid this, we continue sampling until all epsiodes
    # are complete.
    #
    # To start with, all environments are active.
    active = np.ones(venv.num_envs, dtype=bool)
    state = None
    dones = np.zeros(venv.num_envs, dtype=bool)
    while np.any(active):
        # policy gets unwrapped observations (eg as dict, not dictobs)
        acts, state = get_actions(obs, state, dones)
        obs, rews, dones, infos = venv.step(acts)
        assert isinstance(
            obs,
            (np.ndarray, dict),
        ), "Tuple observations are not supported."
        wrapped_obs = types.maybe_wrap_in_dictobs(obs)

        # If an environment is inactive, i.e. the episode completed for that
        # environment after `sample_until(trajectories)` was true, then we do
        # *not* want to add any subsequent trajectories from it. We avoid this
        # by just making it never done.
        dones &= active

        new_trajs = trajectories_accum.add_steps_and_auto_finish(
            acts,
            wrapped_obs,
            rews,
            dones,
            infos,
        )
        trajectories.extend(new_trajs)

        if sample_until(trajectories):
            # Termination condition has been reached. Mark as inactive any
            # environments where a trajectory was completed this timestep.
            active &= ~dones

    # Note that we just drop partial trajectories. This is not ideal for some
    # algos; e.g. BC can probably benefit from partial trajectories, too.

    # Each trajectory is sampled i.i.d.; however, shorter episodes are added to
    # `trajectories` sooner. Shuffle to avoid bias in order. This is important
    # when callees end up truncating the number of trajectories or transitions.
    # It is also cheap, since we're just shuffling pointers.
    rng.shuffle(trajectories)  # type: ignore[arg-type]

    # Sanity checks.
    for trajectory in trajectories:
        n_steps = len(trajectory.acts)
        # extra 1 for the end
        if isinstance(venv.observation_space, spaces.Dict):
            exp_obs = {}
            for k, v in venv.observation_space.items():
                assert v.shape is not None
                exp_obs[k] = (n_steps + 1,) + v.shape
        else:
            obs_space_shape = venv.observation_space.shape
            assert obs_space_shape is not None
            exp_obs = (n_steps + 1,) + obs_space_shape  # type: ignore[assignment]
        real_obs = trajectory.obs.shape
        assert real_obs == exp_obs, f"expected shape {exp_obs}, got {real_obs}"
        assert venv.action_space.shape is not None
        exp_act = (n_steps,) + venv.action_space.shape
        real_act = trajectory.acts.shape
        assert real_act == exp_act, f"expected shape {exp_act}, got {real_act}"
        exp_rew = (n_steps,)
        real_rew = trajectory.rews.shape
        assert real_rew == exp_rew, f"expected shape {exp_rew}, got {real_rew}"

    return trajectories


def rollout_stats(
    trajectories: Sequence[types.TrajectoryWithRew],
) -> Mapping[str, float]:
    """Calculates various stats for a sequence of trajectories.

    Args:
        trajectories: Sequence of trajectories.

    Returns:
        Dictionary containing `n_traj` collected (int), along with episode return
        statistics (keys: `{monitor_,}return_{min,mean,std,max}`, float values)
        and trajectory length statistics (keys: `len_{min,mean,std,max}`, float
        values).

        `return_*` values are calculated from environment rewards.
        `monitor_*` values are calculated from Monitor-captured rewards, and
        are only included if the `trajectories` contain Monitor infos.
    """
    assert len(trajectories) > 0
    out_stats: Dict[str, float] = {"n_traj": len(trajectories)}
    traj_descriptors = {
        "return": np.asarray([sum(t.rews) for t in trajectories]),
        "len": np.asarray([len(t.rews) for t in trajectories]),
    }

    monitor_ep_returns = []
    for t in trajectories:
        if t.infos is not None:
            ep_return = t.infos[-1].get("episode", {}).get("r")
            if ep_return is not None:
                monitor_ep_returns.append(ep_return)
    if monitor_ep_returns:
        # Note monitor_ep_returns[i] may be from a different episode than ep_return[i]
        # since we skip episodes with None infos. This is OK as we only return summary
        # statistics, but you cannot e.g. compute the correlation between ep_return and
        # monitor_ep_returns.
        traj_descriptors["monitor_return"] = np.asarray(monitor_ep_returns)
        # monitor_return_len may be < n_traj when infos is sometimes missing
        out_stats["monitor_return_len"] = len(traj_descriptors["monitor_return"])

    stat_names = ["min", "mean", "std", "max"]
    for desc_name, desc_vals in traj_descriptors.items():
        for stat_name in stat_names:
            stat_value: np.generic = getattr(np, stat_name)(desc_vals)
            # Convert numpy type to float or int. The numpy operators always return
            # a numpy type, but we want to return type float. (int satisfies
            # float type for the purposes of static-typing).
            out_stats[f"{desc_name}_{stat_name}"] = stat_value.item()

    for v in out_stats.values():
        assert isinstance(v, (int, float))
    return out_stats


def flatten_trajectories(
    trajectories: Iterable[types.Trajectory],
) -> types.Transitions:
    """Flatten a series of trajectory dictionaries into arrays.

    Args:
        trajectories: list of trajectories.

    Returns:
        The trajectories flattened into a single batch of Transitions.
    """

    def all_of_type(key, desired_type):
        return all(
            isinstance(getattr(traj, key), desired_type) for traj in trajectories
        )

    assert all_of_type("obs", types.DictObs) or all_of_type("obs", np.ndarray)
    assert all_of_type("acts", np.ndarray)

    # mypy struggles without Any annotation here.
    # The necessary constraints are enforced above.
    keys = ["obs", "next_obs", "acts", "dones", "infos"]
    parts: Mapping[str, List[Any]] = {key: [] for key in keys}
    for traj in trajectories:
        parts["acts"].append(traj.acts)

        obs = traj.obs
        parts["obs"].append(obs[:-1])
        parts["next_obs"].append(obs[1:])

        dones = np.zeros(len(traj.acts), dtype=bool)
        dones[-1] = traj.terminal
        parts["dones"].append(dones)

        if traj.infos is None:
            infos = np.array([{}] * len(traj))
        else:
            infos = traj.infos
        parts["infos"].append(infos)

    cat_parts = {
        key: types.concatenate_maybe_dictobs(part_list)
        for key, part_list in parts.items()
    }
    lengths = set(map(len, cat_parts.values()))
    assert len(lengths) == 1, f"expected one length, got {lengths}"
    return types.Transitions(**cat_parts)


def flatten_trajectories_with_rew(
    trajectories: Sequence[types.TrajectoryWithRew],
) -> types.TransitionsWithRew:
    transitions = flatten_trajectories(trajectories)
    rews = np.concatenate([traj.rews for traj in trajectories])
    return types.TransitionsWithRew(
        **types.dataclass_quick_asdict(transitions),
        rews=rews,
    )


def generate_transitions(
    policy: AnyPolicy,
    venv: VecEnv,
    n_timesteps: int,
    rng: np.random.Generator,
    *,
    truncate: bool = True,
    **kwargs: Any,
) -> types.TransitionsWithRew:
    """Generate obs-action-next_obs-reward tuples.

    Args:
        policy: Can be any of the following:
            - A stable_baselines3 policy or algorithm trained on the gym environment
            - A Callable that takes an ndarray of observations and returns an ndarray
            of corresponding actions
            - None, in which case actions will be sampled randomly
        venv: The vectorized environments to interact with.
        n_timesteps: The minimum number of timesteps to sample.
        rng: The random state to use for sampling trajectories.
        truncate: If True, then drop any additional samples to ensure that exactly
            `n_timesteps` samples are returned.
        **kwargs: Passed-through to generate_trajectories.

    Returns:
        A batch of Transitions. The length of the constituent arrays is guaranteed
        to be at least `n_timesteps` (if specified), but may be greater unless
        `truncate` is provided as we collect data until the end of each episode.
    """
    traj = generate_trajectories(
        policy,
        venv,
        sample_until=make_min_timesteps(n_timesteps),
        rng=rng,
        **kwargs,
    )
    transitions = flatten_trajectories_with_rew(traj)
    if truncate and n_timesteps is not None:
        as_dict = types.dataclass_quick_asdict(transitions)
        truncated = {k: arr[:n_timesteps] for k, arr in as_dict.items()}
        transitions = types.TransitionsWithRew(**truncated)
    return transitions


def rollout(
    policy: AnyPolicy,
    venv: VecEnv,
    sample_until: GenTrajTerminationFn,
    rng: np.random.Generator,
    *,
    unwrap: bool = True,
    exclude_infos: bool = True,
    verbose: bool = True,
    **kwargs: Any,
) -> Sequence[types.TrajectoryWithRew]:
    """Generate policy rollouts.

    This method is a wrapper of generate_trajectories that allows
    the user to additionally replace the rewards and observations with the original
    values if the environment is wrapped, to exclude the infos from the
    trajectories, and to print summary statistics of the rollout.

    The `.infos` field of each Trajectory is set to `None` to save space.

    Args:
        policy: Can be any of the following:
            1) A stable_baselines3 policy or algorithm trained on the gym environment.
            2) A Callable that takes an ndarray of observations and returns an ndarray
            of corresponding actions.
            3) None, in which case actions will be sampled randomly.
        venv: The vectorized environments.
        sample_until: End condition for rollout sampling.
        rng: Random state to use for sampling.
        unwrap: If True, then save original observations and rewards (instead of
            potentially wrapped observations and rewards) by calling
            `unwrap_traj()`.
        exclude_infos: If True, then exclude `infos` from pickle by setting
            this field to None. Excluding `infos` can save a lot of space during
            pickles.
        verbose: If True, then print out rollout stats before saving.
        **kwargs: Passed through to `generate_trajectories`.

    Returns:
        Sequence of trajectories, satisfying `sample_until`. Additional trajectories
        may be collected to avoid biasing process towards short episodes; the user
        should truncate if required.
    """
    trajs = generate_trajectories(
        policy,
        venv,
        sample_until,
        rng=rng,
        **kwargs,
    )
    if unwrap:
        trajs = [unwrap_traj(traj) for traj in trajs]
    if exclude_infos:
        trajs = [dataclasses.replace(traj, infos=None) for traj in trajs]
    if verbose:
        stats = rollout_stats(trajs)
        logging.info(f"Rollout stats: {stats}")
    return trajs


def discounted_sum(arr: np.ndarray, gamma: float) -> Union[np.ndarray, float]:
    """Calculate the discounted sum of `arr`.

    If `arr` is an array of rewards, then this computes the return;
    however, it can also be used to e.g. compute discounted state
    occupancy measures.

    Args:
        arr: 1 or 2-dimensional array to compute discounted sum over.
            Last axis is timestep, from current time step (first) to
            last timestep (last). First axis (if present) is batch
            dimension.
        gamma: the discount factor used.

    Returns:
        The discounted sum over the timestep axis. The first timestep is undiscounted,
        i.e. we start at gamma^0.
    """
    # We want to calculate sum_{t = 0}^T gamma^t r_t, which can be
    # interpreted as the polynomial sum_{t = 0}^T r_t x^t
    # evaluated at x=gamma.
    # Compared to first computing all the powers of gamma, then
    # multiplying with the `arr` values and then summing, this method
    # should require fewer computations and potentially be more
    # numerically stable.
    assert arr.ndim in (1, 2)
    if gamma == 1.0:
        return arr.sum(axis=0)
    else:
        return np.polynomial.polynomial.polyval(gamma, arr)



"""Serialization utilities for trajectories."""
import logging
import os
import warnings
from typing import Mapping, Sequence, cast

import datasets
import numpy as np

from imitation.data import huggingface_utils
from imitation.data.types import AnyPath, Trajectory, TrajectoryWithRew
from imitation.util import util


def save(path: AnyPath, trajectories: Sequence[Trajectory]) -> None:
    """Save a sequence of Trajectories to disk using HuggingFace's datasets library.

    Args:
        path: Trajectories are saved to this path.
        trajectories: The trajectories to save.
    """
    p = util.parse_path(path)
    huggingface_utils.trajectories_to_dataset(trajectories).save_to_disk(str(p))
    logging.info(f"Dumped demonstrations to {p}.")


def load(path: AnyPath) -> Sequence[Trajectory]:
    """Loads a sequence of trajectories saved by `save()` from `path`."""
    # Interestingly, np.load will just silently load a normal pickle file when you
    # set `allow_pickle=True`. So this call should succeed for both the new compressed
    # .npz format and the old pickle based format. To tell the difference, we need to
    # look at the type of the resulting object. If it's the new compressed format,
    # it should be a Mapping that we need to decode, whereas if it's the old format,
    # it's just the sequence of trajectories, and we can return it directly.

    if os.path.isdir(path):  # huggingface datasets format
        dataset = datasets.load_from_disk(str(path))
        if not isinstance(dataset, datasets.Dataset):  # pragma: no cover
            raise ValueError(
                f"Expected to load a `datasets.Dataset` but got {type(dataset)}",
            )

        return huggingface_utils.TrajectoryDatasetSequence(dataset)

    data = np.load(path, allow_pickle=True)  # works for both .npz and .pkl

    if isinstance(data, Sequence):  # pickle format
        warnings.warn("Loading old pickle version of Trajectories", DeprecationWarning)
        return data
    if isinstance(data, Mapping):  # .npz format
        warnings.warn("Loading old npz version of Trajectories", DeprecationWarning)
        num_trajs = len(data["indices"])
        fields = [
            # Account for the extra obs in each trajectory
            np.split(data["obs"], data["indices"] + np.arange(num_trajs) + 1),
            np.split(data["acts"], data["indices"]),
            np.split(data["infos"], data["indices"]),
            data["terminal"],
        ]
        if "rews" in data:
            fields = [
                *fields,
                np.split(data["rews"], data["indices"]),
            ]
            return [TrajectoryWithRew(*args) for args in zip(*fields)]
        else:
            return [Trajectory(*args) for args in zip(*fields)]  # pragma: no cover
    else:  # pragma: no cover
        raise ValueError(
            f"Expected either an .npz file or a pickled sequence of trajectories; "
            f"got a pickled object of type {type(data).__name__}",
        )


def load_with_rewards(path: AnyPath) -> Sequence[TrajectoryWithRew]:
    """Loads a sequence of trajectories with rewards from a file."""
    data = load(path)

    mismatched_types = [
        type(traj) for traj in data if not isinstance(traj, TrajectoryWithRew)
    ]
    if mismatched_types:
        raise ValueError(
            f"Expected all trajectories to be of type `TrajectoryWithRew`, "
            f"but found {mismatched_types[0].__name__}",
        )

    return cast(Sequence[TrajectoryWithRew], data)



"""Tests `imitation.policies.*`."""

import functools
from typing import cast

import gymnasium as gym
import numpy as np
import pytest
import torch as th
from stable_baselines3.common import preprocessing
from torch import nn

from imitation.data import rollout
from imitation.policies import base, serialize
from imitation.util import registry, util

SIMPLE_DISCRETE_ENV = "CartPole-v0"  # Discrete(2) action space
SIMPLE_CONTINUOUS_ENV = "MountainCarContinuous-v0"  # Box(1) action space
SIMPLE_ENVS = [SIMPLE_DISCRETE_ENV, SIMPLE_CONTINUOUS_ENV]
NONTRAINABLE_TYPES = ["random", "zero"]

assert_equal = functools.partial(th.testing.assert_close, rtol=0, atol=0)


@pytest.mark.parametrize("env_name", SIMPLE_ENVS)
@pytest.mark.parametrize("policy_type", NONTRAINABLE_TYPES)
def test_actions_valid(env_name, policy_type, rng):
    """Test output actions of our custom policies always lie in action space."""
    venv = util.make_vec_env(
        env_name,
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    policy = serialize.load_policy(policy_type, venv)
    transitions = rollout.generate_transitions(
        policy,
        venv,
        n_timesteps=100,
        rng=rng,
    )

    for a in transitions.acts:
        assert venv.action_space.contains(a)


@pytest.mark.parametrize(
    "policy_env_name_pair",
    [
        ("ppo", SIMPLE_DISCRETE_ENV),
        ("sac", SIMPLE_CONTINUOUS_ENV),
    ],
)
def test_save_stable_model_errors_and_warnings(
    tmpdir,
    policy_env_name_pair,
    rng,
):
    """Check errors and warnings in `save_stable_model()`."""
    policy, env_name = policy_env_name_pair
    tmpdir = util.parse_path(tmpdir)
    venv = util.make_vec_env(env_name, rng=rng)

    # Trigger FileNotFoundError for no model.{zip,pkl}
    dir_a = tmpdir / "a"
    dir_a.mkdir()
    with pytest.raises(FileNotFoundError, match=".*Expected.*model.zip.*"):
        serialize.load_policy(policy, venv, path=str(dir_a))

    (dir_a / "vec_normalize.pkl").touch()
    (dir_a / "model.zip").touch()
    with pytest.raises(FileExistsError, match="Outdated policy format.*"):
        serialize.load_policy(policy, venv, path=str(dir_a))

    # Trigger FileNotError for nonexistent directory
    dir_nonexistent = tmpdir / "i_dont_exist"
    with pytest.raises(FileNotFoundError):
        serialize.load_policy(policy, venv, path=str(dir_nonexistent))


def _test_serialize_identity(env_name, model_cfg, tmpdir, rng):
    """Test output actions of deserialized policy are same as original."""
    venv = util.make_vec_env(
        env_name,
        n_envs=1,
        parallel=False,
        rng=rng,
    )

    model_name, model_cls_name = model_cfg
    model_cls = registry.load_attr(model_cls_name)

    model = model_cls("MlpPolicy", venv)
    model.learn(1000)

    venv.seed(0)
    venv.reset()
    orig_rollout = rollout.generate_transitions(
        model,
        venv,
        n_timesteps=1000,
        deterministic_policy=True,
        rng=np.random.default_rng(0),
    )

    serialize.save_stable_model(util.parse_path(tmpdir), model)
    loaded = serialize.load_policy(model_name, venv, path=tmpdir)
    venv.seed(0)
    venv.reset()
    new_rollout = rollout.generate_transitions(
        loaded,
        venv,
        n_timesteps=1000,
        deterministic_policy=True,
        rng=np.random.default_rng(0),
    )

    assert np.allclose(orig_rollout.acts, new_rollout.acts)


SB_CONFIGS = serialize.STABLE_BASELINES_CLASSES.items()
CONTINUOUS_ONLY = ["sac"]
NORMAL_CONFIGS = [cfg for cfg in SB_CONFIGS if cfg[0] not in CONTINUOUS_ONLY]
CONTINUOUS_ONLY_CONFIGS = [cfg for cfg in SB_CONFIGS if cfg[0] in CONTINUOUS_ONLY]


@pytest.mark.parametrize("env_name", SIMPLE_ENVS)
@pytest.mark.parametrize("model_cfg", NORMAL_CONFIGS)
def test_serialize_identity(env_name, model_cfg, tmpdir, rng):
    """Test output actions of deserialized policy are same as original."""
    _test_serialize_identity(env_name, model_cfg, tmpdir, rng)


@pytest.mark.parametrize("env_name", [SIMPLE_CONTINUOUS_ENV])
@pytest.mark.parametrize("model_cfg", CONTINUOUS_ONLY_CONFIGS)
def test_serialize_identity_continuous_only(
    env_name,
    model_cfg,
    tmpdir,
    rng,
):
    """Test serialize identity for continuous_only algorithms."""
    _test_serialize_identity(env_name, model_cfg, tmpdir, rng)


class ZeroModule(nn.Module):
    """Module that always returns zeros of same shape as input."""

    def __init__(self, features_dim: int):
        """Builds ZeroModule."""
        super().__init__()
        self.features_dim = features_dim

    def forward(self, x: th.Tensor) -> th.Tensor:
        """Returns zeros of same shape as `x`."""
        assert x.shape[1:] == (self.features_dim,)
        return th.zeros_like(x)


@pytest.mark.parametrize(
    "obs_space",
    [
        gym.spaces.Box(-1, 1, shape=(1,)),
        gym.spaces.Box(-1, 1, shape=(3,)),
        gym.spaces.Box(-1, 1, shape=(3, 4)),
        gym.spaces.Discrete(2),
        gym.spaces.MultiDiscrete([5, 2]),
    ],
)
def test_normalize_features_extractor(obs_space: gym.Space) -> None:
    """Tests `base.NormalizeFeaturesExtractor`.

    Verifies it returns features of appropriate shape (based on flattening `obs_space).
    Also checks that using an identity normalizer leaves observations unchanged,
    and that using `ZeroModule` returns all zeros.

    Args:
        obs_space: The observation space to sample from.
    """
    extractors = {
        "norm": base.NormalizeFeaturesExtractor(obs_space),
        "identity": base.NormalizeFeaturesExtractor(obs_space, nn.Identity),
        "zero": base.NormalizeFeaturesExtractor(obs_space, ZeroModule),
    }

    for i in range(10):
        obs = th.as_tensor([obs_space.sample()])
        # TODO(juan) the cast below is because preprocess_obs has too general a type.
        #  this should be replaced with an overload or a generic.
        #  https://github.com/DLR-RM/stable-baselines3/issues/1065
        obs = cast(th.Tensor, preprocessing.preprocess_obs(obs, obs_space))
        assert isinstance(obs, th.Tensor)
        flattened_obs = obs.flatten(1, -1)
        extracted = {k: extractor(obs) for k, extractor in extractors.items()}
        for k, v in extracted.items():
            assert v.shape == flattened_obs.shape, k

        assert_equal(extracted["identity"], flattened_obs)
        assert_equal(extracted["zero"], flattened_obs * 0.0)



"""This ingredient provides a vectorized gym environment."""
import contextlib
from typing import Any, Generator, Mapping

import numpy as np
import sacred
from stable_baselines3.common import vec_env

from imitation.data import wrappers
from imitation.util import util

environment_ingredient = sacred.Ingredient("environment")


@environment_ingredient.config
def config():
    num_vec = 8  # number of environments in VecEnv
    parallel = True  # Use SubprocVecEnv rather than DummyVecEnv
    max_episode_steps = None  # Set to positive int to limit episode horizons
    env_make_kwargs = {}  # The kwargs passed to `spec.make`.
    gym_id = "seals/CartPole-v0"  # The environment to train on

    locals()  # quieten flake8


@contextlib.contextmanager
@environment_ingredient.capture
def make_venv(
    gym_id: str,
    num_vec: int,
    parallel: bool,
    max_episode_steps: int,
    env_make_kwargs: Mapping[str, Any],
    _run: sacred.run.Run,
    _rnd: np.random.Generator,
    **kwargs,
) -> Generator[vec_env.VecEnv, None, None]:
    """Builds the vector environment.

    Args:
        gym_id: The id of the environment to create.
        num_vec: Number of `gym.Env` instances to combine into a vector environment.
        parallel: Whether to use "true" parallelism. If True, then use `SubProcVecEnv`.
            Otherwise, use `DummyVecEnv` which steps through environments serially.
        max_episode_steps: If not None, then a TimeLimit wrapper is applied to each
            environment to artificially limit the maximum number of timesteps in an
            episode.
        env_make_kwargs: The kwargs passed to `spec.make` of a gym environment.
        kwargs: Passed through to `util.make_vec_env`.

    Yields:
        The constructed vector environment.
    """
    # Note: we create the venv outside the try -- finally block for the case that env
    #     creation fails.
    venv = util.make_vec_env(
        gym_id,
        rng=_rnd,
        n_envs=num_vec,
        parallel=parallel,
        max_episode_steps=max_episode_steps,
        log_dir=_run.config["logging"]["log_dir"] if "logging" in _run.config else None,
        env_make_kwargs=env_make_kwargs,
        **kwargs,
    )
    try:
        yield venv
    finally:
        venv.close()


@contextlib.contextmanager
@environment_ingredient.capture
def make_rollout_venv(
    gym_id: str,
    num_vec: int,
    parallel: bool,
    max_episode_steps: int,
    env_make_kwargs: Mapping[str, Any],
    _rnd: np.random.Generator,
) -> Generator[vec_env.VecEnv, None, None]:
    """Builds the vector environment for rollouts.

    This environment does no logging, and it is wrapped in a `RolloutInfoWrapper`.

    Args:
        gym_id: The id of the environment to create.
        num_vec: Number of `gym.Env` instances to combine into a vector environment.
        parallel: Whether to use "true" parallelism. If True, then use `SubProcVecEnv`.
            Otherwise, use `DummyVecEnv` which steps through environments serially.
        max_episode_steps: If not None, then a TimeLimit wrapper is applied to each
            environment to artificially limit the maximum number of timesteps in an
            episode.
        env_make_kwargs: The kwargs passed to `spec.make` of a gym environment.
        _rnd: Random number generator provided by Sacred.

    Yields:
        The constructed vector environment.
    """
    # Note: we create the venv outside the try -- finally block for the case that env
    #     creation fails.
    venv = util.make_vec_env(
        gym_id,
        rng=_rnd,
        n_envs=num_vec,
        parallel=parallel,
        max_episode_steps=max_episode_steps,
        log_dir=None,
        env_make_kwargs=env_make_kwargs,
        post_wrappers=[lambda env, i: wrappers.RolloutInfoWrapper(env)],
    )
    try:
        yield venv
    finally:
        venv.close()


@environment_ingredient.named_config
def fast():
    num_vec = 2
    parallel = False  # easier to debug with everything in one process
    max_episode_steps = 5

    locals()  # quieten flake8



"""Adversarial Inverse Reinforcement Learning (AIRL)."""
from typing import Optional

import torch as th
from stable_baselines3.common import base_class, policies, vec_env
from stable_baselines3.sac import policies as sac_policies

from imitation.algorithms import base
from imitation.algorithms.adversarial import common
from imitation.rewards import reward_nets

STOCHASTIC_POLICIES = (sac_policies.SACPolicy, policies.ActorCriticPolicy)


class AIRL(common.AdversarialTrainer):
    """Adversarial Inverse Reinforcement Learning (`AIRL`_).

    .. _AIRL: https://arxiv.org/abs/1710.11248
    """

    def __init__(
        self,
        *,
        demonstrations: base.AnyTransitions,
        demo_batch_size: int,
        venv: vec_env.VecEnv,
        gen_algo: base_class.BaseAlgorithm,
        reward_net: reward_nets.RewardNet,
        **kwargs,
    ):
        """Builds an AIRL trainer.

        Args:
            demonstrations: Demonstrations from an expert (optional). Transitions
                expressed directly as a `types.TransitionsMinimal` object, a sequence
                of trajectories, or an iterable of transition batches (mappings from
                keywords to arrays containing observations, etc).
            demo_batch_size: The number of samples in each batch of expert data. The
                discriminator batch size is twice this number because each discriminator
                batch contains a generator sample for every expert sample.
            venv: The vectorized environment to train in.
            gen_algo: The generator RL algorithm that is trained to maximize
                discriminator confusion. Environment and logger will be set to
                `venv` and `custom_logger`.
            reward_net: Reward network; used as part of AIRL discriminator.
            **kwargs: Passed through to `AdversarialTrainer.__init__`.

        Raises:
            TypeError: If `gen_algo.policy` does not have an `evaluate_actions`
                attribute (present in `ActorCriticPolicy`), needed to compute
                log-probability of actions.
        """
        super().__init__(
            demonstrations=demonstrations,
            demo_batch_size=demo_batch_size,
            venv=venv,
            gen_algo=gen_algo,
            reward_net=reward_net,
            **kwargs,
        )
        # AIRL needs a policy from STOCHASTIC_POLICIES to compute discriminator output.
        if not isinstance(self.gen_algo.policy, STOCHASTIC_POLICIES):
            raise TypeError(
                "AIRL needs a stochastic policy to compute the discriminator output.",
            )

    def logits_expert_is_high(
        self,
        state: th.Tensor,
        action: th.Tensor,
        next_state: th.Tensor,
        done: th.Tensor,
        log_policy_act_prob: Optional[th.Tensor] = None,
    ) -> th.Tensor:
        r"""Compute the discriminator's logits for each state-action sample.

        In Fu's AIRL paper (https://arxiv.org/pdf/1710.11248.pdf), the
        discriminator output was given as

        .. math::

            D_{\theta}(s,a) =
            \frac{ \exp{r_{\theta}(s,a)} } { \exp{r_{\theta}(s,a)} + \pi(a|s) }

        with a high value corresponding to the expert and a low value corresponding to
        the generator.

        In other words, the discriminator output is the probability that the action is
        taken by the expert rather than the generator.

        The logit of the above is given as

        .. math::

            \operatorname{logit}(D_{\theta}(s,a)) = r_{\theta}(s,a) - \log{ \pi(a|s) }

        which is what is returned by this function.

        Args:
            state: The state of the environment at the time of the action.
            action: The action taken by the expert or generator.
            next_state: The state of the environment after the action.
            done: whether a `terminal state` (as defined under the MDP of the task) has
                been reached.
            log_policy_act_prob: The log probability of the action taken by the
                generator, :math:`\log{ \pi(a|s) }`.

        Returns:
            The logits of the discriminator for each state-action sample.

        Raises:
            TypeError: If `log_policy_act_prob` is None.
        """
        if log_policy_act_prob is None:
            raise TypeError(
                "Non-None `log_policy_act_prob` is required for this method.",
            )
        reward_output_train = self._reward_net(state, action, next_state, done)
        return reward_output_train - log_policy_act_prob

    @property
    def reward_train(self) -> reward_nets.RewardNet:
        return self._reward_net

    @property
    def reward_test(self) -> reward_nets.RewardNet:
        """Returns the unshaped version of reward network used for testing."""
        reward_net = self._reward_net
        # Recursively return the base network of the wrapped reward net
        while isinstance(reward_net, reward_nets.RewardNetWrapper):
            reward_net = reward_net.base
        return reward_net



"""Helper methods for the `sacred` experimental configuration and logging framework."""

import json
import os
import pathlib
import warnings
from typing import Any, Callable, NamedTuple, Optional, Sequence

import sacred
import sacred.observers
import sacred.run

from imitation.data import types
from imitation.util import util


class SacredDicts(NamedTuple):
    """Each dict `foo` is loaded from `f"{sacred_dir}/foo.json"`."""

    sacred_dir: pathlib.Path
    config: dict
    run: dict

    @classmethod
    def load_from_dir(cls, sacred_dir: pathlib.Path):
        return cls(
            sacred_dir=sacred_dir,
            config=json.loads((sacred_dir / "config.json").read_text()),
            run=json.loads((sacred_dir / "run.json").read_text()),
        )


def dir_contains_sacred_jsons(dir_path: pathlib.Path) -> bool:
    run_path = dir_path / "run.json"
    config_path = dir_path / "config.json"
    return run_path.is_file() and config_path.is_file()


def filter_subdirs(
    root_dir: pathlib.Path,
    filter_fn: Callable[[pathlib.Path], bool] = dir_contains_sacred_jsons,
    *,
    nested_ok: bool = False,
) -> Sequence[pathlib.Path]:
    """Walks through a directory tree, returning paths to filtered subdirectories.

    Does not follow symlinks.

    Args:
        root_dir: The start of the directory tree walk.
        filter_fn: A function with takes a directory path and returns True if
            we should include the directory path in this function's return value.
        nested_ok: Allow returning "nested" directories, i.e. a return value where
            some elements are subdirectories of other elements.

    Returns:
        A list of all subdirectory paths where `filter_fn(path) == True`.

    Raises:
        ValueError: If `nested_ok` is False and one of the filtered directory
            paths is a subdirecotry of another.
    """
    filtered_dirs = set()
    for root_str, _, _ in os.walk(root_dir, followlinks=False):
        root = pathlib.Path(root_str)
        if filter_fn(root):
            filtered_dirs.add(root)

    if not nested_ok:
        for dirpath in filtered_dirs:
            for other_dirpath in filtered_dirs:
                if dirpath != other_dirpath and other_dirpath in dirpath.parents:
                    raise ValueError(
                        f"Found nested directories: {dirpath} and {other_dirpath}",
                    )
    return list(filtered_dirs)


def build_sacred_symlink(log_dir: types.AnyPath, run: sacred.run.Run) -> None:
    """Constructs a symlink "{log_dir}/sacred" => "${SACRED_PATH}"."""
    log_dir = util.parse_path(log_dir)

    sacred_dir = get_sacred_dir_from_run(run)
    if sacred_dir is None:
        warnings.warn(RuntimeWarning("Couldn't find sacred directory."))
        return
    symlink_path = log_dir / "sacred"
    target_path = pathlib.Path(os.path.relpath(sacred_dir, start=log_dir))

    # Path.symlink_to errors if the symlink already exists. In our case, we actually
    # want to override the symlink to point to the most recent Sacred dir. The
    # examples/quickstart.sh script fails without this check when run a second time.
    #
    # If `symlink_path` exists and is not a symlink, then it was created by something
    # other than this function then we don't remove it (and will error on the symlink
    # step).
    if symlink_path.is_symlink():
        symlink_path.unlink()

    # Use relative paths so we can mount the output directory at different paths
    # (e.g. when copying across machines).
    try:
        symlink_path.symlink_to(target_path, target_is_directory=True)
    except OSError as e:
        if os.name == "nt":  # Windows
            msg = (
                "Exception occurred while creating symlink. "
                "Please ensure that Developer mode is enabled."
            )
            raise OSError(msg) from e
        else:
            raise e


def get_sacred_dir_from_run(run: sacred.run.Run) -> Optional[pathlib.Path]:
    """Returns path to the sacred directory, or None if not found."""
    for obs in run.observers:
        if isinstance(obs, sacred.observers.FileStorageObserver):
            return util.parse_path(obs.dir)
    return None


def dict_get_nested(d: dict, nested_key: str, *, sep=".", default=None) -> Any:
    curr = d
    for key in nested_key.split(sep):
        if isinstance(curr, dict) and key in curr:
            curr = curr[key]
        else:
            return default
    return curr



"""Tunes the hyperparameters of the algorithms."""

import copy
import pathlib
from typing import Dict

import numpy as np
import ray
from pandas.api import types as pd_types
from ray.tune.search import optuna
from sacred.observers import FileStorageObserver

from imitation.scripts.config.parallel import parallel_ex
from imitation.scripts.config.tuning import tuning_ex


@tuning_ex.main
def tune(
    parallel_run_config,
    eval_best_trial_resource_multiplier: int = 1,
    num_eval_seeds: int = 5,
) -> None:
    """Tune hyperparameters of imitation algorithms using the parallel script.

    The parallel script is called twice in this function. The first call is to
    tune the hyperparameters. The second call is to evaluate the best trial on
    a separate set of seeds.

    Args:
        parallel_run_config: Dictionary of arguments to pass to the parallel script.
            This is used to define the search space for tuning the hyperparameters.
        eval_best_trial_resource_multiplier: Factor by which to multiply the
            number of cpus per trial in `resources_per_trial`. This is useful for
            allocating more resources per trial to the evaluation trials than the
            resources for hyperparameter tuning since number of evaluation trials
            is usually much smaller than the number of tuning trials.
        num_eval_seeds: Number of distinct seeds to evaluate the best trial on.
            Set to 0 to disable evaluation.

    Raises:
        ValueError: If no trials are returned by the parallel run of tuning.
    """
    updated_parallel_run_config = copy.deepcopy(parallel_run_config)
    search_alg = optuna.OptunaSearch()
    tune_run_kwargs = updated_parallel_run_config.setdefault("tune_run_kwargs", dict())
    tune_run_kwargs["search_alg"] = search_alg
    run = parallel_ex.run(config_updates=updated_parallel_run_config)
    experiment_analysis = run.result
    if not experiment_analysis.trials:
        raise ValueError(
            "No trials found. Please ensure that the `experiment_checkpoint_path` "
            "in `parallel_run_config` is passed correctly "
            "or that the tuning run finished properly.",
        )

    return_key = "imit_stats/monitor_return_mean"
    if updated_parallel_run_config["sacred_ex_name"] == "train_rl":
        return_key = "monitor_return_mean"
    best_trial = find_best_trial(experiment_analysis, return_key, print_return=True)

    if num_eval_seeds > 0:  # evaluate the best trial
        resources_per_trial_eval = copy.deepcopy(
            updated_parallel_run_config["resources_per_trial"],
        )
        # update cpus per trial only if it is provided in `resources_per_trial`
        # Uses the default values (cpu=1) if it is not provided
        if "cpu" in updated_parallel_run_config["resources_per_trial"]:
            resources_per_trial_eval["cpu"] *= eval_best_trial_resource_multiplier
        evaluate_trial(
            best_trial,
            num_eval_seeds,
            updated_parallel_run_config["run_name"] + "_best_hp_eval",
            updated_parallel_run_config,
            resources_per_trial_eval,
            return_key,
        )


def find_best_trial(
    experiment_analysis: ray.tune.analysis.ExperimentAnalysis,
    return_key: str,
    print_return: bool = False,
) -> ray.tune.experiment.Trial:
    """Find the trial with the best mean return across all seeds.

    Args:
        experiment_analysis: The result of a parallel/tuning experiment.
        return_key: The key of the return metric in the results dataframe.
        print_return: Whether to print the mean and std of the returns
            of the best trial.

    Returns:
        best_trial: The trial with the best mean return across all seeds.
    """
    df = experiment_analysis.results_df
    # convert object dtype to str required by df.groupby
    for col in df.columns:
        if pd_types.is_object_dtype(df[col]):
            df[col] = df[col].astype("str")
    # group into separate HP configs
    grp_keys = [c for c in df.columns if c.startswith("config")]
    grp_keys = [c for c in grp_keys if "seed" not in c and "trial_index" not in c]
    grps = df.groupby(grp_keys)
    # store mean return of runs across all seeds in a group
    # the transform method is applied to get the mean return for every trial
    # instead of for every group. So every trial in a group will have the same
    # mean return column.
    df["mean_return"] = grps[return_key].transform(lambda x: x.mean())
    best_config_df = df[df["mean_return"] == df["mean_return"].max()]
    row = best_config_df.iloc[0]
    best_config_tag = row["experiment_tag"]
    assert experiment_analysis.trials is not None  # for mypy
    best_trial = [
        t for t in experiment_analysis.trials if best_config_tag in t.experiment_tag
    ][0]

    if print_return:
        all_returns = df[df["mean_return"] == row["mean_return"]][return_key]
        all_returns = all_returns.to_numpy()
        print("All returns:", all_returns)
        print("Mean return:", row["mean_return"])
        print("Std return:", np.std(all_returns))
        print("Total seeds:", len(all_returns))
    return best_trial


def evaluate_trial(
    trial: ray.tune.experiment.Trial,
    num_eval_seeds: int,
    run_name: str,
    parallel_run_config,
    resources_per_trial: Dict[str, int],
    return_key: str,
    print_return: bool = False,
):
    """Evaluate a given trial of a parallel run on a separate set of seeds.

    Args:
        trial: The trial to evaluate.
        num_eval_seeds: Number of distinct seeds to evaluate the best trial on.
        run_name: The name of the evaluation run.
        parallel_run_config: Dictionary of arguments passed to the parallel
            script to get best_trial.
        resources_per_trial: Resources to be used for each evaluation trial.
        return_key: The key of the return metric in the results dataframe.
        print_return: Whether to print the mean and std of the evaluation returns.

    Returns:
        eval_run: The result of the evaluation run.
    """
    config = trial.config
    config["config_updates"].update(
        seed=ray.tune.grid_search(list(range(100, 100 + num_eval_seeds))),
    )
    eval_config_updates = parallel_run_config.copy()
    eval_config_updates.update(
        run_name=run_name,
        num_samples=1,
        search_space=config,
        resources_per_trial=resources_per_trial,
        repeat=1,
        experiment_checkpoint_path="",
    )
    # required for grid search
    eval_config_updates["tune_run_kwargs"].update(search_alg=None)
    eval_run = parallel_ex.run(config_updates=eval_config_updates)
    eval_result = eval_run.result
    returns = eval_result.results_df[return_key].to_numpy()
    if print_return:
        print("All returns:", returns)
        print("Mean:", np.mean(returns))
        print("Std:", np.std(returns))
    return eval_run


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "tuning"
    observer = FileStorageObserver(observer_path)
    tuning_ex.observers.append(observer)
    tuning_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Test imitation.policies.replay_buffer_wrapper."""

import os.path as osp
from typing import Type

import numpy as np
import pytest
import stable_baselines3 as sb3
import torch as th
from stable_baselines3.common import buffers, off_policy_algorithm, policies
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.save_util import load_from_pkl

from imitation.policies.replay_buffer_wrapper import ReplayBufferRewardWrapper
from imitation.util import util


def zero_reward_fn(
    state: np.ndarray,
    action: np.ndarray,
    next_state: np.ndarray,
    done: np.ndarray,
) -> np.ndarray:
    del action, next_state, done
    return np.zeros(state.shape[0], dtype=np.float32)


def make_algo_with_wrapped_buffer(
    rl_cls: Type[off_policy_algorithm.OffPolicyAlgorithm],
    policy_cls: Type[BasePolicy],
    replay_buffer_class: Type[buffers.ReplayBuffer],
    rng: np.random.Generator,
    buffer_size: int = 100,
) -> off_policy_algorithm.OffPolicyAlgorithm:
    venv = util.make_vec_env("Pendulum-v1", n_envs=1, rng=rng)
    rl_algo = rl_cls(
        policy=policy_cls,
        policy_kwargs=dict(),
        env=venv,
        seed=42,
        replay_buffer_class=ReplayBufferRewardWrapper,
        replay_buffer_kwargs=dict(
            replay_buffer_class=replay_buffer_class,
            reward_fn=zero_reward_fn,
        ),
        buffer_size=buffer_size,
    )  # type: ignore[call-arg]
    return rl_algo


def test_invalid_args(rng):
    with pytest.raises(
        TypeError,
        match=r".*unexpected keyword argument 'replay_buffer_class'.*",
    ):
        # we ignore the type because we are intentionally
        # passing the wrong type for the test
        make_algo_with_wrapped_buffer(
            rl_cls=sb3.PPO,  # type: ignore[arg-type]
            policy_cls=policies.ActorCriticPolicy,
            replay_buffer_class=buffers.ReplayBuffer,
            rng=rng,
        )

    with pytest.raises(AssertionError, match=r".*only ReplayBuffer is supported.*"):
        make_algo_with_wrapped_buffer(
            rl_cls=sb3.SAC,
            policy_cls=sb3.sac.policies.SACPolicy,
            replay_buffer_class=buffers.DictReplayBuffer,
            rng=rng,
        )


def test_wrapper_class(tmpdir, rng):
    buffer_size = 15
    total_timesteps = 20

    rl_algo = make_algo_with_wrapped_buffer(
        rl_cls=sb3.SAC,
        policy_cls=sb3.sac.policies.SACPolicy,
        replay_buffer_class=buffers.ReplayBuffer,
        buffer_size=buffer_size,
        rng=rng,
    )

    rl_algo.learn(total_timesteps=total_timesteps)

    buffer_path = osp.join(tmpdir, "buffer.pkl")
    rl_algo.save_replay_buffer(buffer_path)
    replay_buffer_wrapper = load_from_pkl(buffer_path)
    replay_buffer = replay_buffer_wrapper.replay_buffer

    # replay_buffer_wrapper.sample(...) should return zero-reward transitions
    assert buffer_size == replay_buffer_wrapper.size() == replay_buffer.size()
    assert (replay_buffer_wrapper.sample(total_timesteps).rewards == 0.0).all()
    assert (replay_buffer.sample(total_timesteps).rewards != 0.0).all()  # seed=42

    # replay_buffer_wrapper.pos, replay_buffer_wrapper.full
    assert replay_buffer_wrapper.pos == total_timesteps - buffer_size
    assert replay_buffer_wrapper.full

    # reset()
    replay_buffer_wrapper.reset()
    assert 0 == replay_buffer_wrapper.size() == replay_buffer.size()
    assert replay_buffer_wrapper.pos == 0
    assert not replay_buffer_wrapper.full

    # to_torch()
    tensor = replay_buffer_wrapper.to_torch(np.ones(42))
    assert type(tensor) is th.Tensor

    # raise error for _get_samples()
    with pytest.raises(NotImplementedError, match=r".*_get_samples.*"):
        replay_buffer_wrapper._get_samples()



"""Runs a Sacred experiment in parallel."""

import collections.abc
import copy
import pathlib
from typing import Any, Callable, Dict, Mapping, Sequence

import ray
import ray.tune
import sacred
from ray.tune import search
from ray.tune.search import optuna
from sacred.observers import FileStorageObserver

from imitation.scripts.config.parallel import parallel_ex


@parallel_ex.main
def parallel(
    sacred_ex_name: str,
    run_name: str,
    num_samples: int,
    search_space: Mapping[str, Any],
    base_named_configs: Sequence[str],
    base_config_updates: Mapping[str, Any],
    resources_per_trial: Mapping[str, Any],
    init_kwargs: Mapping[str, Any],
    repeat: int,
    experiment_checkpoint_path: str,
    tune_run_kwargs: Dict[str, Any],
) -> ray.tune.ExperimentAnalysis:
    """Parallelize multiple runs of another Sacred Experiment using Ray Tune.

    A Sacred FileObserver is attached to the inner experiment and writes Sacred
    logs to "{RAY_LOCAL_DIR}/sacred/". These files are automatically copied over
    to `upload_dir` if that argument is provided in `tune_run_kwargs`.

    Args:
        sacred_ex_name: The Sacred experiment to tune. Either "train_rl",
            "train_imitation", "train_adversarial" or "train_preference_comparisons".
        run_name: A name describing this parallelizing experiment.
            This argument is also passed to `ray.tune.run` as the `name` argument.
            It is also saved in 'sacred/run.json' of each inner Sacred experiment
            under the 'experiment.name' key. This is equivalent to using the Sacred
            CLI '--name' option on the inner experiment. Offline analysis jobs can use
            this argument to group similar data.
        num_samples: Number of times to sample from the hyperparameter space without
            considering repetition using `repeat`.
        search_space: A dictionary which can contain Ray Tune search objects like
            `ray.tune.grid_search` and `ray.tune.sample_from`, and is
            passed as the `config` argument to `ray.tune.run()`. After the
            `search_space` is transformed by Ray, it passed into
            `sacred_ex.run(**run_kwargs)` as `run_kwargs` (`sacred_ex` is the Sacred
            Experiment selected via `sacred_ex_name`). Usually `search_space` only has
            the keys "named_configs" and "config_updates", but any parameter names
            to `sacred.Experiment.run()` are okay.
        base_named_configs: Default Sacred named configs. Any named configs
            taken from `search_space` are higher priority than the base_named_configs.
            Concretely, this priority is implemented by appending named configs taken
            from `search_space` to the run's named configs after `base_named_configs`.
            Named configs in `base_named_configs` don't appear in the automatically
            generated Ray directory name, unlike named configs from `search_space`.
        base_config_updates: Default Sacred config updates. Any config updates taken
            from `search_space` are higher priority than `base_config_updates`.
            Config updates in `base_config_updates` don't appear in the automatically
            generated Ray directory name, unlike config updates from `search_space`.
        resources_per_trial: Argument to `ray.tune.run()`.
        init_kwargs: Arguments to pass to `ray.init`.
        repeat: Number of runs to repeat each trial for.
            If `repeat` > 1, then optuna is used as the default search algorithm
            unless specified otherwise in `tune_run_kwargs`.
        experiment_checkpoint_path: Path containing the checkpoints of a previous
            experiment ran using this script. Useful for evaluating the best trial
            of the experiment.
        tune_run_kwargs: Other arguments to pass to `ray.tune.run()`.

    Raises:
        TypeError: Named configs not string sequences or config updates not mappings.
        ValueError: `repeat` > 1 but `search_alg` is not an instance of
            `ray.tune.search.SearchAlgorithm`.

    Returns:
        The result of running the parallel experiment with `ray.tune.run()`.
        Useful for fetching the configs and results dataframe of all the trials.
    """
    # Basic validation for config options before we enter parallel jobs.
    if not isinstance(base_named_configs, collections.abc.Sequence):
        raise TypeError("base_named_configs must be a Sequence")

    if not isinstance(base_config_updates, collections.abc.Mapping):
        raise TypeError("base_config_updates must be a Mapping")

    if not isinstance(search_space["named_configs"], collections.abc.Sequence):
        raise TypeError('search_space["named_configs"] must be a Sequence')

    if not isinstance(search_space["config_updates"], collections.abc.Mapping):
        raise TypeError('search_space["config_updates"] must be a Mapping')

    # Convert Sacred's ReadOnlyList to List because not picklable.
    base_named_configs = list(base_named_configs)

    # Convert Sacred's ReadOnlyDict (and recursively convert ReadOnlyContainer values)
    # to regular python variants because not picklable.
    base_config_updates = copy.deepcopy(base_config_updates)
    search_space = copy.deepcopy(search_space)

    trainable = _ray_tune_sacred_wrapper(
        sacred_ex_name,
        run_name,
        base_named_configs,
        base_config_updates,
    )

    ray.init(**init_kwargs)
    updated_tune_run_kwargs = copy.deepcopy(tune_run_kwargs)
    if repeat > 1:
        try:
            # Use optuna as the default search algorithm for repeat runs.
            algo = tune_run_kwargs.get("search_alg", optuna.OptunaSearch())
            updated_tune_run_kwargs["search_alg"] = search.Repeater(algo, repeat)
        except AttributeError as e:
            raise ValueError(
                "repeat > 1 but search_alg is not an instance of "
                "ray.tune.search.SearchAlgorithm",
            ) from e

    if sacred_ex_name == "train_rl":
        return_key = "monitor_return_mean"
    else:
        return_key = "imit_stats/monitor_return_mean"

    try:
        if experiment_checkpoint_path:
            result = ray.tune.ExperimentAnalysis(experiment_checkpoint_path)
        else:
            result = ray.tune.run(
                trainable,
                config=search_space,
                num_samples=num_samples * repeat,
                name=run_name,
                resources_per_trial=resources_per_trial,
                metric=return_key,
                mode="max",
                **updated_tune_run_kwargs,
            )
        return result
    finally:
        ray.shutdown()


def _ray_tune_sacred_wrapper(
    sacred_ex_name: str,
    run_name: str,
    base_named_configs: list,
    base_config_updates: Mapping[str, Any],
) -> Callable[[Dict[str, Any], Any], Mapping[str, Any]]:
    """From an Experiment build a wrapped run function suitable for Ray Tune.

    `ray.tune.run(...)` expects a trainable function that takes a dict
    argument `config`. The wrapped function uses `config` as keyword args for
    `ex.run(...)` because we want to be able to hyperparameter tune over both the
    `named_configs` and `config_updates` arguments.

    The Ray Tune `reporter` is not passed to the inner experiment.

    Args:
        sacred_ex_name: The Sacred experiment to tune. Either "train_rl" or
            "train_adversarial".
        run_name: A name describing this parallelizing experiment.
            This argument is also passed to `ray.tune.run` as the `name` argument.
            It is also saved in 'sacred/run.json' of each inner Sacred experiment
            under the 'experiment.name' key. This is equivalent to using the Sacred
            CLI '--name' option on the inner experiment. Offline analysis jobs can use
            this argument to group similar data.
        base_named_configs: Default Sacred named configs. Any named configs
            taken from `search_space` are higher priority than the base_named_configs.
            Concretely, this priority is implemented by appending named configs taken
            from `search_space` to the run's named configs after `base_named_configs`.
            Named configs in `base_named_configs` don't appear in the automatically
            generated Ray directory name, unlike named configs from `search_space`.
        base_config_updates: Default Sacred config updates. Any config updates taken
            from `search_space` are higher priority than `base_config_updates`.
            Config updates in `base_config_updates` don't appear in the automatically
            generated Ray directory name, unlike config updates from `search_space`.

    Returns:
        A function that takes two arguments, `config` (used as keyword args for
        `ex.run`) and `reporter`. The function returns the run result.
    """

    def inner(config: Mapping[str, Any], reporter) -> Mapping[str, Any]:
        """Trainable function with the correct signature for `ray.tune`.

        Args:
            config: Keyword arguments for `ex.run()`, where `ex` is the
                `sacred.Experiment` instance associated with `sacred_ex_name`.
            reporter: Callback to report progress to Ray.

        Returns:
            Result from `ray.Run` object.
        """
        # Set Sacred capture mode to "sys" because default "fd" option leads to error.
        # See https://github.com/IDSIA/sacred/issues/289.
        # TODO(shwang): Stop modifying CAPTURE_MODE once the issue is fixed.
        sacred.SETTINGS.CAPTURE_MODE = "sys"

        run_kwargs = dict(**config)
        updated_run_kwargs: Dict[str, Any] = {}
        # Import inside function rather than in module because Sacred experiments
        # are not picklable, and Ray requires this function to be picklable.
        from imitation.scripts.train_adversarial import train_adversarial_ex
        from imitation.scripts.train_imitation import train_imitation_ex
        from imitation.scripts.train_preference_comparisons import (
            train_preference_comparisons_ex,
        )
        from imitation.scripts.train_rl import train_rl_ex

        experiments = {
            "train_rl": train_rl_ex,
            "train_adversarial": train_adversarial_ex,
            "train_imitation": train_imitation_ex,
            "train_preference_comparisons": train_preference_comparisons_ex,
        }
        ex = experiments[sacred_ex_name]

        # Apply base configs to get modified `named_configs` and `config_updates`.
        named_configs = base_named_configs + run_kwargs["named_configs"]
        updated_run_kwargs["named_configs"] = named_configs

        config_updates: Dict[str, Any] = {}
        config_updates.update(base_config_updates)
        config_updates.update(run_kwargs["config_updates"])
        # for repeat runs, set the seed using their trial index
        if "__trial_index__" in run_kwargs:
            config_updates.update(seed=run_kwargs.pop("__trial_index__"))
        updated_run_kwargs["config_updates"] = config_updates

        # Add other run_kwargs items to updated_run_kwargs.
        for k, v in run_kwargs.items():
            if k not in updated_run_kwargs:
                updated_run_kwargs[k] = v
        run = ex.run(
            **updated_run_kwargs,
            options={"--run": run_name, "--file_storage": "sacred"},
        )

        assert run.status == "COMPLETED"
        return run.result

    return inner


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "parallel"
    observer = FileStorageObserver(observer_path)
    parallel_ex.observers.append(observer)
    parallel_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Tests for `imitation.algorithms.dagger`."""

import contextlib
import glob
import math
import os
from typing import List, Optional, Sequence
from unittest import mock

import gymnasium as gym
import numpy as np
import pytest
import torch.random
from stable_baselines3.common import evaluation, policies

from imitation.algorithms import bc, dagger
from imitation.data import rollout, serialize, types
from imitation.data.types import TrajectoryWithRew
from imitation.policies import base
from imitation.testing import reward_improvement
from imitation.util import util


@pytest.fixture(params=[True, False])
def maybe_pendulum_expert_trajectories(
    pendulum_expert_trajectories: Sequence[TrajectoryWithRew],
    request,
) -> Optional[Sequence[TrajectoryWithRew]]:
    keep_trajs = request.param
    if keep_trajs:
        return pendulum_expert_trajectories
    else:
        return None


@pytest.mark.parametrize("num_rampdown_rounds", [1, 2, 3, 10])
def test_linear_beta_schedule(num_rampdown_rounds):
    # GIVEN
    sched = dagger.LinearBetaSchedule(num_rampdown_rounds)
    idx_after_rampdown = num_rampdown_rounds + 1

    # WHEN
    betas = [sched(i) for i in range(num_rampdown_rounds + 10)]

    # THEN
    assert np.allclose(
        betas[:idx_after_rampdown],
        np.linspace(1, 0, idx_after_rampdown),
    )
    assert np.allclose(betas[idx_after_rampdown:], 0)


@pytest.mark.parametrize("decay_probability", [0.1, 0.5, 0.9, 1])
def test_exponential_beta_schedule(decay_probability):
    # GIVEN
    sched = dagger.ExponentialBetaSchedule(decay_probability)

    # WHEN
    betas = [sched(i) for i in range(10)]

    # THEN
    assert np.allclose(betas, decay_probability ** np.arange(10))


@pytest.mark.parametrize("decay_probability", [-0.1, 0, 1.1, 2])
def test_forbidden_decay_probability_on_exp_beta_schedule(decay_probability):
    with pytest.raises(ValueError):
        dagger.ExponentialBetaSchedule(decay_probability)


def test_traj_collector_seed(tmpdir, pendulum_venv, rng):
    collector = dagger.InteractiveTrajectoryCollector(
        venv=pendulum_venv,
        get_robot_acts=lambda o: [
            pendulum_venv.action_space.sample() for _ in range(len(o))
        ],
        beta=0.5,
        save_dir=tmpdir,
        rng=rng,
    )
    seeds1 = collector.seed(42)
    obs1 = collector.reset()
    seeds2 = collector.seed(42)
    obs2 = collector.reset()

    np.testing.assert_array_equal(seeds1, seeds2)
    np.testing.assert_array_equal(obs1, obs2)


def test_traj_collector(tmpdir, pendulum_venv, rng):
    robot_calls = 0
    num_episodes = 0

    def get_random_acts(obs):
        nonlocal robot_calls
        robot_calls += len(obs)
        return [pendulum_venv.action_space.sample() for _ in range(len(obs))]

    collector = dagger.InteractiveTrajectoryCollector(
        venv=pendulum_venv,
        get_robot_acts=get_random_acts,
        beta=0.5,
        save_dir=tmpdir,
        rng=rng,
    )
    collector.reset()
    zero_acts = np.zeros(
        (pendulum_venv.num_envs,) + pendulum_venv.action_space.shape,
        dtype=pendulum_venv.action_space.dtype,
    )
    obs, rews, dones, infos = collector.step(zero_acts)
    assert np.all(rews != 0)
    assert not np.any(dones)
    for info in infos:
        assert isinstance(info, dict)
    # roll out 5 * venv.num_envs episodes (Pendulum-v1 has 200 timestep episodes)
    for _ in range(1000):
        _, _, dones, _ = collector.step(zero_acts)
        num_episodes += np.sum(dones)

    # there is a <10^(-12) probability this fails by chance; we should be calling
    # robot with 50% prob each time
    assert 388 * pendulum_venv.num_envs <= robot_calls <= 612 * pendulum_venv.num_envs

    # All user/expert actions are zero. Therefore, all collected actions should be
    # zero.
    file_paths = glob.glob(os.path.join(tmpdir, "dagger-demo-*.npz"))
    assert num_episodes == 5 * pendulum_venv.num_envs
    assert len(file_paths) == num_episodes
    trajs = [serialize.load(p)[0] for p in file_paths]
    nonzero_acts = sum(np.sum(traj.acts != 0) for traj in trajs)
    assert nonzero_acts == 0


def test_traj_collector_reproducible(tmpdir, pendulum_venv):
    # We run the collector twice with the same random seeds and
    # check that the following 2 properties hold:
    # 1) The files written in the first run have the same filenames
    #    as the files written in the second run.
    # 2) Each file in the first run stores the same trajectory as
    #    the file with the same filename in the second run.
    filename_to_traj_dicts = []
    with torch.random.fork_rng():
        for run_idx in range(2):
            # Reset the random seeds.
            save_dir = os.path.join(tmpdir, "run{0}".format(run_idx))
            rng = np.random.default_rng(12345)
            pendulum_venv.seed(12345)
            pendulum_venv.action_space.seed(12345)

            # Run the collector.
            collector = dagger.InteractiveTrajectoryCollector(
                venv=pendulum_venv,
                get_robot_acts=lambda o: [
                    pendulum_venv.action_space.sample() for _ in range(len(o))
                ],
                beta=0.5,
                save_dir=save_dir,
                rng=rng,
            )
            collector.reset()
            zero_acts = np.zeros(
                (pendulum_venv.num_envs,) + pendulum_venv.action_space.shape,
                dtype=pendulum_venv.action_space.dtype,
            )
            for _ in range(1000):
                _, _, dones, _ = collector.step(zero_acts)

            # Get the observations from all the collected trajectories.
            file_paths = glob.glob(os.path.join(save_dir, "dagger-demo-*.npz"))
            filename_to_traj_dict = {}
            for fp in file_paths:
                traj = serialize.load_with_rewards(fp)[0]
                # For the purposes of testing, we remove `infos` from the
                # trajectory, because `infos` contains the time that it
                # takes to complete an episode, which we expect to differ
                # slightly between runs.
                traj_without_infos = types.TrajectoryWithRew(
                    obs=traj.obs,
                    acts=traj.acts,
                    infos=None,
                    terminal=traj.terminal,
                    rews=traj.rews,
                )
                filename = os.path.basename(fp)
                filename_to_traj_dict[filename] = traj_without_infos
            filename_to_traj_dicts.append(filename_to_traj_dict)

    assert filename_to_traj_dicts[0] == filename_to_traj_dicts[1]


def _build_dagger_trainer(
    tmpdir,
    venv,
    beta_schedule,
    expert_policy,
    pendulum_expert_rollouts: List[TrajectoryWithRew],
    custom_logger,
    rng: np.random.Generator,
):
    del expert_policy
    if pendulum_expert_rollouts is not None:
        pytest.skip(
            "DAggerTrainer does not use trajectories. "
            "Skipping to avoid duplicate test.",
        )
    bc_trainer = bc.BC(
        observation_space=venv.observation_space,
        action_space=venv.action_space,
        optimizer_kwargs=dict(lr=1e-3),
        custom_logger=custom_logger,
        rng=rng,
    )
    return dagger.DAggerTrainer(
        venv=venv,
        scratch_dir=tmpdir,
        beta_schedule=beta_schedule,
        bc_trainer=bc_trainer,
        custom_logger=custom_logger,
        rng=rng,
    )


def _build_simple_dagger_trainer(
    tmpdir,
    venv,
    beta_schedule,
    expert_policy,
    pendulum_expert_rollouts: Optional[List[TrajectoryWithRew]],
    custom_logger,
    rng,
):
    bc_trainer = bc.BC(
        observation_space=venv.observation_space,
        action_space=venv.action_space,
        optimizer_kwargs=dict(lr=1e-3),
        custom_logger=custom_logger,
        rng=rng,
    )
    return dagger.SimpleDAggerTrainer(
        venv=venv,
        scratch_dir=tmpdir,
        beta_schedule=beta_schedule,
        bc_trainer=bc_trainer,
        expert_policy=expert_policy,
        expert_trajs=pendulum_expert_rollouts,
        custom_logger=custom_logger,
        rng=rng,
    )


@pytest.fixture(params=[None, dagger.LinearBetaSchedule(1)])
def beta_schedule(request):
    return request.param


@pytest.fixture(params=[_build_dagger_trainer, _build_simple_dagger_trainer])
def init_trainer_fn(
    request,
    tmpdir,
    pendulum_venv,
    beta_schedule,
    pendulum_expert_policy,
    maybe_pendulum_expert_trajectories: Optional[List[TrajectoryWithRew]],
    custom_logger,
    rng,
):
    # Provide a trainer initialization fixture in addition `trainer` fixture below
    # for tests that want to initialize multiple DAggerTrainer.
    trainer_fn = request.param
    return lambda: trainer_fn(
        tmpdir,
        pendulum_venv,
        beta_schedule,
        pendulum_expert_policy,
        maybe_pendulum_expert_trajectories,
        custom_logger,
        rng,
    )


@pytest.fixture
def trainer(init_trainer_fn):
    return init_trainer_fn()


@pytest.fixture
def simple_dagger_trainer(
    tmpdir,
    pendulum_venv,
    beta_schedule,
    pendulum_expert_policy,
    maybe_pendulum_expert_trajectories: Optional[List[TrajectoryWithRew]],
    custom_logger,
    rng,
):
    return _build_simple_dagger_trainer(
        tmpdir,
        pendulum_venv,
        beta_schedule,
        pendulum_expert_policy,
        maybe_pendulum_expert_trajectories,
        custom_logger,
        rng,
    )


def test_trainer_needs_demos_exception_error(
    trainer,
    maybe_pendulum_expert_trajectories: Optional[List[TrajectoryWithRew]],
):
    assert trainer.round_num == 0
    error_ctx = pytest.raises(dagger.NeedsDemosException)
    ctx: contextlib.AbstractContextManager
    if maybe_pendulum_expert_trajectories is not None and isinstance(
        trainer,
        dagger.SimpleDAggerTrainer,
    ):
        # In this case, demos should be preloaded and we shouldn't experience
        # the NeedsDemoException error.
        ctx = contextlib.nullcontext()
    else:
        # In all cases except the one above, an error should be raised because
        # there are no demos to update on.
        ctx = error_ctx

    with ctx:
        trainer.extend_and_update(dict(n_epochs=1))

    # If ctx==nullcontext before, then we should fail on the second call
    # because there aren't any demos loaded into round 1 yet.
    # If ctx==error_ctx, then still should fail once again on the second call.
    with error_ctx:
        trainer.extend_and_update(dict(n_epochs=1))


def test_trainer_train_arguments(trainer, pendulum_expert_policy, rng):
    def add_samples():
        collector = trainer.create_trajectory_collector()
        rollout.generate_trajectories(
            pendulum_expert_policy,
            collector,
            sample_until=rollout.make_min_timesteps(40),
            rng=rng,
        )

    # Lower default number of epochs for the no-arguments call that follows.
    add_samples()
    with mock.patch.object(trainer, "DEFAULT_N_EPOCHS", 1):
        trainer.extend_and_update()

    add_samples()
    trainer.extend_and_update(dict(n_batches=2))

    add_samples()
    trainer.extend_and_update(dict(n_epochs=1))


def test_trainer_makes_progress(init_trainer_fn, pendulum_venv, pendulum_expert_policy):
    with torch.random.fork_rng():
        # manually seed to avoid flakiness
        torch.random.manual_seed(42)
        pendulum_venv.action_space.seed(42)

        trainer = init_trainer_fn()
        novice_rewards, _ = evaluation.evaluate_policy(
            trainer.policy,
            pendulum_venv,
            25,
            deterministic=True,
            return_episode_rewards=True,
        )
        # note a randomly initialised policy does well for some seeds -- so may
        # want to adjust this check if changing seed. Pendulum return can range
        # from -1,200 to -130 (approx.), per Figure 3 in this PDF (on page 3):
        # https://arxiv.org/pdf/2106.09556.pdf
        assert np.mean(novice_rewards) < -1000
        # Train for 5 iterations. (4 or fewer causes test to fail on some configs.)
        # see https://github.com/HumanCompatibleAI/imitation/issues/580 for details
        for _ in range(5):
            # roll out a few trajectories for dataset, then train for a few steps
            collector = trainer.create_trajectory_collector()
            for _ in range(4):
                obs = collector.reset()
                dones = [False] * pendulum_venv.num_envs
                while not np.any(dones):
                    expert_actions, _ = pendulum_expert_policy.predict(
                        obs,
                        deterministic=True,
                    )
                    obs, _, dones, _ = collector.step(expert_actions)
            trainer.extend_and_update(dict(n_epochs=1))
        # make sure we're doing better than a random policy would
        rewards_after_training, _ = evaluation.evaluate_policy(
            trainer.policy,
            pendulum_venv,
            25,
            deterministic=True,
            return_episode_rewards=True,
        )

    assert reward_improvement.is_significant_reward_improvement(
        novice_rewards,
        rewards_after_training,
    )
    assert reward_improvement.mean_reward_improved_by(
        novice_rewards,
        rewards_after_training,
        300,
    )


@pytest.mark.parametrize(
    "init_trainer_fn",
    [_build_dagger_trainer, _build_simple_dagger_trainer],
)
def test_trainer_reproducible(
    init_trainer_fn,
    tmpdir,
    pendulum_venv,
    pendulum_expert_policy,
    custom_logger,
):
    # Check that we get the same results if we run the trainer
    # twice with the same random seeds.
    # In particular, check that the trajectories from rolling out
    # the trained policy are the same in each run.
    run_trajs = []
    with torch.random.fork_rng():
        for run_idx in range(2):
            # Reset the random seeds.
            run_dir = os.path.join(tmpdir, "run{0}".format(run_idx))
            torch.random.manual_seed(12345)
            rng = np.random.default_rng(12345)
            pendulum_venv.seed(12345)
            pendulum_venv.action_space.seed(12345)

            beta_schedule = None
            maybe_pendulum_expert_trajectories = None
            trainer = init_trainer_fn(
                run_dir,
                pendulum_venv,
                beta_schedule,
                pendulum_expert_policy,
                maybe_pendulum_expert_trajectories,
                custom_logger,
                rng,
            )

            for _ in range(2):
                collector = trainer.create_trajectory_collector()
                obs = collector.reset()
                dones = [False] * pendulum_venv.num_envs
                while not np.any(dones):
                    expert_actions, _ = pendulum_expert_policy.predict(
                        obs,
                        deterministic=True,
                    )
                    obs, _, dones, _ = collector.step(expert_actions)
                trainer.extend_and_update(dict(n_epochs=1))

            trajs = rollout.rollout(
                trainer.policy,
                pendulum_venv,
                rollout.make_sample_until(min_episodes=2),
                rng,
            )
            run_trajs.append(trajs)

    assert len(run_trajs) == 2
    assert run_trajs[0] == run_trajs[1]


def test_trainer_save_reload(tmpdir, init_trainer_fn, pendulum_venv):
    trainer = init_trainer_fn()
    trainer.round_num = 3
    trainer.save_trainer()
    loaded_trainer = dagger.reconstruct_trainer(trainer.scratch_dir, venv=pendulum_venv)
    assert loaded_trainer.round_num == trainer.round_num

    # old trainer and reloaded trainer should have same variable values
    old_vars = trainer.policy.state_dict()
    new_vars = loaded_trainer.policy.state_dict()
    assert len(new_vars) == len(old_vars)
    assert all(values.equal(old_vars[var]) for var, values in new_vars.items())

    # also those values should be different from freshly initialized trainer
    third_trainer = init_trainer_fn()
    third_vars = third_trainer.policy.state_dict()
    assert len(third_vars) == len(old_vars)
    assert not all(values.equal(old_vars[var]) for var, values in third_vars.items())


@pytest.mark.parametrize("num_episodes", [1, 4])
def test_simple_dagger_trainer_train(
    simple_dagger_trainer: dagger.SimpleDAggerTrainer,
    pendulum_venv,
    num_episodes: int,
    tmpdir: str,
):
    episode_length = 200  # for Pendulum-v1
    rollout_min_episodes = 2
    simple_dagger_trainer.train(
        total_timesteps=episode_length * num_episodes,
        bc_train_kwargs=dict(n_batches=10),
        rollout_round_min_episodes=rollout_min_episodes,
        rollout_round_min_timesteps=1,
    )

    episodes_per_round = max(rollout_min_episodes, pendulum_venv.num_envs)
    num_rounds = math.ceil(num_episodes / episodes_per_round)

    round_paths = glob.glob(os.path.join(tmpdir, "demos", "round-*"))
    assert len(round_paths) == num_rounds
    for directory in round_paths:
        file_paths = glob.glob(os.path.join(directory, "dagger-demo-*.npz"))
        assert len(file_paths) == episodes_per_round


def test_policy_save_reload(tmpdir, trainer):
    # just make sure the methods run; we already test them in test_bc.py
    policy_path = os.path.join(tmpdir, "policy.pt")
    util.save_policy(trainer.policy, policy_path)
    pol = bc.reconstruct_policy(policy_path)
    assert isinstance(pol, policies.BasePolicy)


def test_simple_dagger_space_mismatch_error(
    tmpdir,
    pendulum_venv,
    beta_schedule,
    pendulum_expert_policy,
    maybe_pendulum_expert_trajectories: Optional[List[TrajectoryWithRew]],
    custom_logger,
    rng,
):
    class MismatchedSpace(gym.spaces.Space):
        """Dummy space that is not equal to any other space."""

    # Swap out expert_policy.{observation,action}_space with a bad space to
    # elicit space mismatch errors.
    space = MismatchedSpace()
    for space_name in ["observation", "action"]:
        with mock.patch.object(pendulum_expert_policy, f"{space_name}_space", space):
            with pytest.raises(ValueError, match=f"Mismatched {space_name}.*"):
                _build_simple_dagger_trainer(
                    tmpdir,
                    pendulum_venv,
                    beta_schedule,
                    pendulum_expert_policy,
                    maybe_pendulum_expert_trajectories,
                    custom_logger,
                    rng,
                )


def test_dagger_not_enough_transitions_error(tmpdir, custom_logger, rng):
    venv = util.make_vec_env("CartPole-v0", rng=rng)
    # Initialize with large batch size to ensure error down the line.
    bc_trainer = bc.BC(
        observation_space=venv.observation_space,
        action_space=venv.action_space,
        batch_size=100_000,
        custom_logger=custom_logger,
        rng=rng,
    )
    trainer = dagger.DAggerTrainer(
        venv=venv,
        scratch_dir=tmpdir,
        bc_trainer=bc_trainer,
        custom_logger=custom_logger,
        rng=rng,
    )
    collector = trainer.create_trajectory_collector()
    policy = base.RandomPolicy(venv.observation_space, venv.action_space)
    rollout.generate_trajectories(
        policy,
        collector,
        rollout.make_min_episodes(1),
        rng=rng,
    )
    with pytest.raises(ValueError, match="Not enough transitions.*"):
        trainer.extend_and_update()



"""Tests for commands.py + Smoke tests for bash scripts in experiments/."""

import glob
import os
import pathlib
import re
import subprocess
from typing import List

import pytest

SCRIPT_NAMES = (
    "bc_benchmark.sh",
    "dagger_benchmark.sh",
    "benchmark_and_table.sh",
    "imit_benchmark.sh",
    "rollouts_from_policies.sh",
    "transfer_learn_benchmark.sh",
)

THIS_DIR = pathlib.Path(__file__).absolute().parent
BENCHMARKING_DIR = THIS_DIR.parent / "src/imitation/scripts/config/tuned_hps"
EXPERIMENTS_DIR = THIS_DIR.parent / "experiments"
COMMANDS_PY_PATH = EXPERIMENTS_DIR / "commands.py"

EXPECTED_LOCAL_CONFIG_TEMPLATE = f"""python -m imitation.scripts.train_imitation \
dagger --capture=sys --name=run0 --file_storage={{output_dir}}/sacred/\
$USER-cmd-run0-dagger-0-c9420c90 \
with {BENCHMARKING_DIR}/fast_dagger_seals_cartpole.json \
seed=0 logging.log_root={{output_dir}}"""

BENCHMARKING_DIR_SUFFIX = re.sub(r".*/src/", "", str(BENCHMARKING_DIR))
EXPECTED_HOFVARPNIR_CONFIG_TEMPLATE = f"""ctl job run \
--name $USER-cmd-run0-dagger-0-c9420c90 \
--command "python -m imitation.scripts.train_imitation dagger \
--capture=sys --name=run0 --file_storage={{output_dir}}/sacred/\
$USER-cmd-run0-dagger-0-c9420c90 \
with /data/imitation/src/{BENCHMARKING_DIR_SUFFIX}/fast_dagger_seals_cartpole.json \
seed=0 logging.log_root={{output_dir}}" \
--container hacobe/devbox:imitation \
--login --force-pull --never-restart --gpu 0 \
--shared-host-dir-mount /data"""


def _get_benchmarking_path(benchmarking_file):
    return os.path.join(BENCHMARKING_DIR, benchmarking_file)


def _run_commands_from_flags(**kwargs) -> List[str]:
    """Run commands.py with flags derived from the given `kwargs`.

    This is a helper function to reduce boilerplate code in the tests
    for commands.py.

    Each key-value pair in kwargs corresponds to one flag.
    If the value in the key-value is True, then the flag has the form "--key".
    If the value in the key-value is a list, then the flag has the form
    "--key value[0] value[1] ..."
    Otherwise, the flag has the form "--key=value".

    E.g., _run_commands_from_flags(name="baz", seeds=[0, 1], remote=True)
    will execute the following python command:

    python experiments/commands.py --name=baz --seeds 0 1 --remote

    The function will return the parsed output of executing that command.

    Args:
        kwargs: keyword arguments from which to derive flags.

    Returns:
        A list where the ith entry of the list is the ith line printed out
        by the python command.
    """
    flags = []

    # Add some defaults so that most tests can skip specifying these flags.
    if "name" not in kwargs:
        flags.append("--name=run0")

    if "cfg_pattern" not in kwargs:
        cfg_pattern = _get_benchmarking_path("fast_dagger_seals_cartpole.json")
        flags.append("--cfg_pattern=" + cfg_pattern)

    if "output_dir" not in kwargs:
        flags.append("--output_dir=output")

    # Build the flags.
    for key in kwargs:
        if isinstance(kwargs[key], bool) and kwargs[key]:
            flag = f"--{key}"
        elif isinstance(kwargs[key], list):
            value = " ".join(map(str, kwargs[key]))
            flag = f"--{key} {value}"
        else:
            flag = f"--{key}={kwargs[key]}"
        flags.append(flag)

    py_command = f"python {COMMANDS_PY_PATH} " + " ".join(flags)

    completed_process = subprocess.run(
        py_command,
        shell=True,
        capture_output=True,
        stdin=subprocess.DEVNULL,
        check=True,
    )

    # Each line of the formatted stdout is a command.
    formatted_stdout = completed_process.stdout.decode("ascii").strip()
    commands = formatted_stdout.split("\n")

    return commands


def test_commands_local_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    commands = _run_commands_from_flags()
    assert len(commands) == 1
    expected = EXPECTED_LOCAL_CONFIG_TEMPLATE.format(output_dir="output")
    assert commands[0] == expected


def test_commands_local_config_runs(tmpdir):
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    commands = _run_commands_from_flags(output_dir=tmpdir)
    assert len(commands) == 1
    expected = EXPECTED_LOCAL_CONFIG_TEMPLATE.format(output_dir=tmpdir)
    assert commands[0] == expected
    completed_process = subprocess.run(
        commands[0],
        shell=True,
        capture_output=True,
        stdin=subprocess.DEVNULL,
        check=True,
    )
    assert completed_process.returncode == 0
    assert (tmpdir / "dagger" / "seals-CartPole-v0").exists()
    assert (tmpdir / "sacred").exists()


def test_commands_local_config_with_custom_flags():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    commands = _run_commands_from_flags(
        name="baz",
        seeds=1,
        output_dir="/foo/bar",
    )
    assert len(commands) == 1
    expected = f"""python -m imitation.scripts.train_imitation dagger \
--capture=sys --name=baz --file_storage=/foo/bar/sacred/\
$USER-cmd-baz-dagger-1-c9420c90 \
with {BENCHMARKING_DIR}/fast_dagger_seals_cartpole.json \
seed=1 logging.log_root=/foo/bar"""
    assert commands[0] == expected


def test_commands_hofvarpnir_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    output_dir = "/data/output"
    commands = _run_commands_from_flags(output_dir=output_dir, remote=True)
    assert len(commands) == 1
    expected = EXPECTED_HOFVARPNIR_CONFIG_TEMPLATE.format(output_dir=output_dir)
    assert commands[0] == expected


def test_commands_hofvarpnir_config_with_custom_flags():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    commands = _run_commands_from_flags(
        name="baz",
        remote_cfg_dir="/bas/bat",
        seeds=1,
        output_dir="/foo/bar",
        container="bam",
        remote=True,
    )
    assert len(commands) == 1
    expected = """ctl job run --name $USER-cmd-baz-dagger-1-c9420c90 \
--command "python -m imitation.scripts.train_imitation dagger \
--capture=sys --name=baz --file_storage=/foo/bar/sacred/\
$USER-cmd-baz-dagger-1-c9420c90 \
with /bas/bat/fast_dagger_seals_cartpole.json \
seed=1 logging.log_root=/foo/bar" --container bam \
--login --force-pull --never-restart --gpu 0 \
--shared-host-dir-mount /data"""
    assert commands[0] == expected


def test_commands_local_config_with_special_characters_in_flags(tmpdir):
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    # Simulate running commands.py with the following flag:
    # --output_dir="\"/tmp/.../foo bar\""
    # And generating a training command with the following flag:
    # --logging.log_root="/tmp/.../foo bar"
    #
    # If we didn't enclose the directory in quotes as in:
    # --output_dir=/tmp/.../foo bar
    # Then we would get an "unrecognized arguments: bar" error
    # trying to run commands.py.
    #
    # Or if we enclosed the directory in double quotes as in:
    # --output_dir="/tmp/.../foo bar"
    # Then we would generate a training command with the following flag:
    # --logging.log_root=/tmp/.../foo bar
    # And we would get an error trying to run the generated command.
    output_subdir = "foo bar"
    unquoted_output_dir = (tmpdir / f"{output_subdir}").strpath
    output_dir = '"\\"' + unquoted_output_dir + '\\""'
    commands = _run_commands_from_flags(output_dir=output_dir)
    assert len(commands) == 1
    # The extra double quotes are removed in the generated command.
    # It has the following flag:
    # --logging.log_root="/tmp/.../foo bar"
    # So it can run without an error introduced by the space.
    expected = EXPECTED_LOCAL_CONFIG_TEMPLATE.format(
        output_dir='"' + unquoted_output_dir + '"',
    )
    assert commands[0] == expected


def test_commands_hofvarpnir_config_with_special_characters_in_flags(tmpdir):
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    # See the comments in
    # test_commands_local_config_with_special_characters_in_flags
    # for a discussion of special characters in the flag values.
    output_subdir = "foo bar"
    unquoted_output_dir = (tmpdir / f"{output_subdir}").strpath
    output_dir = '"\\"' + unquoted_output_dir + '\\""'
    commands = _run_commands_from_flags(output_dir=output_dir, remote=True)
    assert len(commands) == 1
    # Make sure double quotes are escaped in the training script command
    # because the training script command is itself enclosed in double
    # quotes within the cluster command.
    expected = EXPECTED_HOFVARPNIR_CONFIG_TEMPLATE.format(
        output_dir='\\"' + unquoted_output_dir + '\\"',
    )
    assert commands[0] == expected


def test_commands_bc_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    cfg_pattern = _get_benchmarking_path("bc_seals_ant_best_hp_eval.json")
    commands = _run_commands_from_flags(cfg_pattern=cfg_pattern)
    assert len(commands) == 1
    expected = f"""python -m imitation.scripts.train_imitation bc \
--capture=sys --name=run0 --file_storage=output/sacred/\
$USER-cmd-run0-bc-0-bb460c12 \
with {BENCHMARKING_DIR}/bc_seals_ant_best_hp_eval.json \
seed=0 logging.log_root=output"""
    assert commands[0] == expected


def test_commands_dagger_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    cfg_pattern = _get_benchmarking_path("dagger_seals_ant_best_hp_eval.json")
    commands = _run_commands_from_flags(cfg_pattern=cfg_pattern)
    assert len(commands) == 1
    expected = f"""python -m imitation.scripts.train_imitation dagger \
--capture=sys --name=run0 --file_storage=output/sacred/\
$USER-cmd-run0-dagger-0-f0790db7 \
with {BENCHMARKING_DIR}/dagger_seals_ant_best_hp_eval.json \
seed=0 logging.log_root=output"""
    assert commands[0] == expected


def test_commands_gail_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    cfg_pattern = _get_benchmarking_path("gail_seals_ant_best_hp_eval.json")
    commands = _run_commands_from_flags(cfg_pattern=cfg_pattern)
    assert len(commands) == 1
    expected = f"""python -m imitation.scripts.train_adversarial gail \
--capture=sys --name=run0 --file_storage=output/sacred/\
$USER-cmd-run0-gail-0-d5be0cea \
with {BENCHMARKING_DIR}/gail_seals_ant_best_hp_eval.json \
seed=0 logging.log_root=output"""
    assert commands[0] == expected


def test_commands_airl_config():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    cfg_pattern = _get_benchmarking_path("airl_seals_ant_best_hp_eval.json")
    commands = _run_commands_from_flags(cfg_pattern=cfg_pattern)
    assert len(commands) == 1
    expected = f"""python -m imitation.scripts.train_adversarial airl \
--capture=sys --name=run0 \
--file_storage=output/sacred/$USER-cmd-run0-airl-0-d7040cf5 \
with {BENCHMARKING_DIR}/airl_seals_ant_best_hp_eval.json \
seed=0 logging.log_root=output"""
    assert commands[0] == expected


def test_commands_multiple_configs():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    # Test a more complicated `cfg_pattern`.
    cfg_pattern = _get_benchmarking_path("*.json")
    commands = _run_commands_from_flags(cfg_pattern=cfg_pattern)
    assert len(commands) == len(glob.glob(cfg_pattern))


def test_commands_multiple_configs_multiple_seeds():
    if os.name == "nt":  # pragma: no cover
        pytest.skip("commands.py not ported to Windows.")
    cfg_pattern = _get_benchmarking_path("*.json")
    seeds = [0, 1, 2]
    commands = _run_commands_from_flags(
        cfg_pattern=cfg_pattern,
        seeds=seeds,
    )
    n_configs = len(glob.glob(cfg_pattern))
    n_seeds = len(seeds)
    assert len(commands) == (n_configs * n_seeds)


@pytest.mark.parametrize(
    "script_name",
    SCRIPT_NAMES,
)
def test_experiments_fast(script_name: str):
    """Quickly check that experiments run successfully on fast mode."""
    if os.name == "nt":  # pragma: no cover
        pytest.skip("bash shell scripts not ported to Windows.")
    exit_code = subprocess.call([f"./{EXPERIMENTS_DIR.stem}/{script_name}", "--fast"])
    assert exit_code == 0



"""This ingredient provides Weights & Biases logging."""

import logging
from typing import Any, Mapping, Optional

import sacred

wandb_ingredient = sacred.Ingredient("logging.wandb")
logger = logging.getLogger(__name__)


@wandb_ingredient.config
def wandb_config():
    # Other users can overwrite this function to customize their wandb.init() call.
    wandb_tag = None  # User-specified tag for this run
    wandb_name_prefix = ""  # User-specified prefix for the run name
    wandb_kwargs = dict(
        project="imitation",
        monitor_gym=False,
        save_code=False,
    )  # Other kwargs to pass to wandb.init()
    wandb_additional_info = dict()

    locals()


@wandb_ingredient.capture
def wandb_init(
    _run,
    wandb_name_prefix: str,
    wandb_tag: Optional[str],
    wandb_kwargs: Mapping[str, Any],
    wandb_additional_info: Mapping[str, Any],
    log_dir: str,
) -> None:
    """Putting everything together to get the W&B kwargs for wandb.init().

    Args:
        wandb_name_prefix: User-specified prefix for wandb run name.
        wandb_tag: User-specified tag for this run.
        wandb_kwargs: User-specified kwargs for wandb.init().
        wandb_additional_info: User-specific additional info to add to wandb experiment
            ``config``.
        log_dir: W&B logs will be stored in directory `{log_dir}/wandb/`.

    Raises:
        ModuleNotFoundError: wandb is not installed.
    """
    env_name = _run.config["environment"]["gym_id"]
    root_seed = _run.config["seed"]

    updated_wandb_kwargs: Mapping[str, Any] = {
        **wandb_kwargs,
        "name": f"{wandb_name_prefix}-{env_name}-seed{root_seed}",
        "tags": [env_name, f"seed{root_seed}"] + ([wandb_tag] if wandb_tag else []),
        "dir": log_dir,
    }
    try:
        import wandb
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Trying to call `wandb.init()` but `wandb` not installed: "
            "try `pip install wandb`.",
        ) from e
    wandb_config_dict = dict(**_run.config)
    wandb_config_dict.update(wandb_additional_info)
    wandb.init(config=wandb_config_dict, **updated_wandb_kwargs)



"""Configuration settings for train_dagger, training DAgger from synthetic demos."""

import pathlib

import sacred

from imitation.scripts.ingredients import bc
from imitation.scripts.ingredients import demonstrations as demos_common
from imitation.scripts.ingredients import environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, sqil

train_imitation_ex = sacred.Experiment(
    "train_imitation",
    ingredients=[
        logging_ingredient.logging_ingredient,
        demos_common.demonstrations_ingredient,
        expert.expert_ingredient,
        environment.environment_ingredient,
        policy_evaluation.policy_evaluation_ingredient,
        bc.bc_ingredient,
        sqil.sqil_ingredient,
    ],
)


@train_imitation_ex.config
def config():
    dagger = dict(
        use_offline_rollouts=False,  # warm-start policy with BC from offline demos
        total_timesteps=1e5,
        beta_schedule=None,
    )


@train_imitation_ex.named_config
def mountain_car():
    environment = dict(gym_id="MountainCar-v0")
    bc = dict(l2_weight=0.0)
    dagger = dict(total_timesteps=20000)


@train_imitation_ex.named_config
def seals_mountain_car():
    environment = dict(gym_id="seals/MountainCar-v0")
    bc = dict(l2_weight=0.0)
    dagger = dict(total_timesteps=20000)


@train_imitation_ex.named_config
def cartpole():
    environment = dict(gym_id="CartPole-v1")
    dagger = dict(total_timesteps=20000)


@train_imitation_ex.named_config
def seals_cartpole():
    environment = dict(gym_id="seals/CartPole-v0")
    dagger = dict(total_timesteps=20000)


@train_imitation_ex.named_config
def pendulum():
    environment = dict(gym_id="Pendulum-v1")


@train_imitation_ex.named_config
def ant():
    environment = dict(gym_id="Ant-v2")


@train_imitation_ex.named_config
def half_cheetah():
    environment = dict(gym_id="HalfCheetah-v2")
    bc = dict(l2_weight=0.0)
    dagger = dict(total_timesteps=60000)


@train_imitation_ex.named_config
def humanoid():
    environment = dict(gym_id="Humanoid-v2")


@train_imitation_ex.named_config
def seals_humanoid():
    environment = dict(gym_id="seals/Humanoid-v0")


@train_imitation_ex.named_config
def fast():
    dagger = dict(total_timesteps=50)
    bc = dict(train_kwargs=dict(n_batches=50))
    sqil = dict(total_timesteps=50)


hyperparam_dir = pathlib.Path(__file__).absolute().parent / "tuned_hps"
tuned_alg_envs = [
    "bc_seals_ant",
    "bc_seals_half_cheetah",
    "bc_seals_hopper",
    "bc_seals_swimmer",
    "bc_seals_walker",
    "dagger_seals_ant",
    "dagger_seals_half_cheetah",
    "dagger_seals_hopper",
    "dagger_seals_swimmer",
    "dagger_seals_walker",
]

for tuned_alg_env in tuned_alg_envs:
    config_file = hyperparam_dir / f"{tuned_alg_env}_best_hp_eval.json"
    assert config_file.is_file(), f"{config_file} does not exist"
    train_imitation_ex.add_named_config(tuned_alg_env, str(config_file))



"""Config files for parallel experiments.

Parallel experiments are intended to be defined in Python rather than
via CLI. For example, a user should add a new
`@parallel_ex.named_config` to define a new parallel experiment.

Adding custom named configs is necessary because the CLI interface can't add
search spaces to the config like `"seed": tune.choice([0, 1, 2, 3])`.

For tuning hyperparameters of an algorithm on a given environment,
check out the imitation/scripts/tuning.py script.
"""

import numpy as np
import ray.tune as tune
import sacred

from imitation.util.util import make_unique_timestamp

parallel_ex = sacred.Experiment("parallel")


@parallel_ex.config
def config():
    sacred_ex_name = "train_rl"  # The experiment to parallelize
    init_kwargs = {}  # Keyword arguments to pass to ray.init()
    _uuid = make_unique_timestamp()
    run_name = f"DEFAULT_{_uuid}"  # CLI --name option. For analysis grouping.
    resources_per_trial = {}  # Argument to `tune.run`
    base_named_configs = []  # Background settings before search_space is applied
    base_config_updates = {}  # Background settings before search_space is applied
    search_space = {
        "named_configs": [],
        "config_updates": {},
    }  # `config` argument to `ray.tune.run(trainable, config)`

    num_samples = 1  # Number of samples per grid search configuration
    repeat = 1  # Number of times to repeat a sampled configuration
    experiment_checkpoint_path = ""  # Path to checkpoint of experiment
    tune_run_kwargs = {}  # Additional kwargs to pass to `tune.run`


# Debug named configs


@parallel_ex.named_config
def generate_test_data():
    """Used by tests/generate_test_data.sh to generate tests/testdata/gather_tb/.

    "tests/testdata/gather_tb/" should contain 2 Tensorboard run directories,
    one for each of the trials in the search space below.
    """
    sacred_ex_name = "train_rl"
    run_name = "TEST"
    repeat = 1
    search_space = {
        "config_updates": {
            "rl": {
                "rl_kwargs": {
                    "learning_rate": tune.choice(
                        [3e-4 * x for x in (1 / 3, 1 / 2)],
                    ),
                },
            },
        },
    }
    base_named_configs = [
        "cartpole",
        "environment.fast",
        "policy_evaluation.fast",
        "rl.fast",
        "fast",
    ]
    base_config_updates = {
        "rollout_save_final": False,
    }


@parallel_ex.named_config
def example_cartpole_rl():
    sacred_ex_name = "train_rl"
    run_name = "example-cartpole"
    repeat = 2
    search_space = {
        "config_updates": {
            "rl": {
                "rl_kwargs": {
                    "learning_rate": tune.choice(np.logspace(3e-6, 1e-1, num=3)),
                    "nminibatches": tune.choice([16, 32, 64]),
                },
            },
        },
    }
    base_named_configs = ["cartpole"]
    resources_per_trial = dict(cpu=4)



"""Tests for `imitation.util.registry`."""

import pytest

from imitation.util import registry


def test_lazy():
    """Test indirect/lazy loading of registered values."""
    reg = registry.Registry()

    reg.register("nomodule", indirect="this.module.does.not.exist:foobar")
    with pytest.raises(ImportError):
        reg.get("nomodule")

    reg.register("noattribute", indirect="imitation:attr_does_not_exist")
    with pytest.raises(AttributeError):
        reg.get("noattribute")

    with pytest.raises(ValueError, match="exactly one of"):
        reg.register(key="wrongargs", value=3.14, indirect="math:pi")

    reg.register("exists", indirect="math:pi")
    val = reg.get("exists")
    import math

    assert val == math.pi


def test_keys():
    reg = registry.Registry()

    with pytest.raises(KeyError, match="not registered"):
        reg.get("foobar")

    reg.register(key="foobar", value="fizzbuzz")
    assert reg.get("foobar") == "fizzbuzz"

    with pytest.raises(KeyError, match="Duplicate registration"):
        reg.register(key="foobar", value="fizzbuzz")



"""Load serialized policies of different types."""

# FIXME(sam): it seems like this module could mostly be replaced with a few
# torch.load() and torch.save() calls

import logging
import pathlib
from typing import Callable, Type, TypeVar

import huggingface_sb3 as hfsb3
from stable_baselines3.common import base_class, callbacks, policies, vec_env

from imitation.policies import base
from imitation.util import registry, util

Algorithm = TypeVar("Algorithm", bound=base_class.BaseAlgorithm)

# Note: a VecEnv will always be passed first and then any kwargs. There is just no
# proper way to specify this in python yet. For details see
# https://stackoverflow.com/questions/61569324/type-annotation-for-callable-that-takes-kwargs
# TODO(juan) this can be fixed using ParamSpec
#  (https://github.com/HumanCompatibleAI/imitation/issues/574)
PolicyLoaderFn = Callable[..., policies.BasePolicy]
"""A policy loader function that takes a VecEnv before any other custom arguments and
returns a stable_baselines3 base policy policy."""

policy_registry: registry.Registry[PolicyLoaderFn] = registry.Registry()
"""Registry of policy loading functions. Add your own here if desired."""


def load_stable_baselines_model(
    cls: Type[Algorithm],
    path: str,
    venv: vec_env.VecEnv,
    **kwargs,
) -> Algorithm:
    """Helper method to load RL models from Stable Baselines.

    Args:
        cls: Stable Baselines RL algorithm.
        path: Path to zip file containing saved model data or to a folder containing a
            `model.zip` file.
        venv: Environment to train on.
        kwargs: Passed through to `cls.load`.

    Raises:
        FileNotFoundError: If `path` is not a directory containing a `model.zip` file.
        FileExistsError: If `path` contains a `vec_normalize.pkl` file (unsupported).

    Returns:
        The deserialized RL algorithm.
    """
    logging.info(f"Loading Stable Baselines policy for '{cls}' from '{path}'")
    path_obj = util.parse_path(path)

    if path_obj.is_dir():
        path_obj = path_obj / "model.zip"
        if not path_obj.exists():
            raise FileNotFoundError(
                f"Expected '{path}' to be a directory containing a 'model.zip' file.",
            )

    # SOMEDAY(adam): added 2022-01, can probably remove this check in 2023
    vec_normalize_path = path_obj.parent / "vec_normalize.pkl"
    if vec_normalize_path.exists():
        raise FileExistsError(
            "Outdated policy format: we do not support restoring normalization "
            f"statistics from '{vec_normalize_path}'",
        )

    return cls.load(path_obj, env=venv, **kwargs)


def _load_stable_baselines_from_file(
    cls: Type[base_class.BaseAlgorithm],
) -> PolicyLoaderFn:
    """Creates a policy loading function to read a policy from a file.

    Args:
        cls: The RL algorithm, e.g. `stable_baselines3.PPO`.

    Returns:
        A function loading policies trained via cls.
    """

    def f(venv: vec_env.VecEnv, path: str) -> policies.BasePolicy:
        """Loads a policy saved to path, for environment env."""
        model = load_stable_baselines_model(cls, path, venv)
        return getattr(model, "policy")

    return f


def _load_stable_baselines_from_huggingface(
    algo_name: str,
    cls: Type[base_class.BaseAlgorithm],
) -> PolicyLoaderFn:
    """Creates a policy loading function to load from Hugging Face.

    Args:
        algo_name: The name of the algorithm, e.g. `ppo`.
        cls: The RL algorithm, e.g. `stable_baselines3.PPO`.

    Returns:
        A function loading policies trained via cls.
    """

    def f(
        venv: vec_env.VecEnv,
        env_name: str,
        organization: str = "HumanCompatibleAI",
    ) -> policies.BasePolicy:
        """Loads a policy saved to path, for environment env."""
        model_name = hfsb3.ModelName(algo_name, hfsb3.EnvironmentName(env_name))
        repo_id = hfsb3.ModelRepoId(organization, model_name)
        filename = hfsb3.load_from_hub(repo_id, model_name.filename)
        model = load_stable_baselines_model(cls, filename, venv)
        return getattr(model, "policy")

    return f


policy_registry.register(
    "random",
    value=registry.build_loader_fn_require_space(base.RandomPolicy),
)
policy_registry.register(
    "zero",
    value=registry.build_loader_fn_require_space(base.ZeroPolicy),
)


def _add_stable_baselines_policies_from_file(classes):
    for k, cls_name in classes.items():
        cls = registry.load_attr(cls_name)
        fn = _load_stable_baselines_from_file(cls)
        policy_registry.register(k, value=fn)


def _add_stable_baselines_policies_from_huggingface(classes):
    for k, cls_name in classes.items():
        cls = registry.load_attr(cls_name)
        fn = _load_stable_baselines_from_huggingface(k, cls)
        policy_registry.register(f"{k}-huggingface", value=fn)


STABLE_BASELINES_CLASSES = {
    "ppo": "stable_baselines3:PPO",
    "sac": "stable_baselines3:SAC",
}
_add_stable_baselines_policies_from_file(STABLE_BASELINES_CLASSES)
_add_stable_baselines_policies_from_huggingface(STABLE_BASELINES_CLASSES)


def load_policy(
    policy_type: str,
    venv: vec_env.VecEnv,
    **kwargs,
) -> policies.BasePolicy:
    """Load serialized policy.

    Note on the kwargs:

    - `zero` and `random` policy take no kwargs
    - `ppo` and `sac` policies take a `path` argument with a path to a zip file or to a
      folder containing a `model.zip` file.
    - `ppo-huggingface` and `sac-huggingface` policies take an `env_name` and optional
      `organization` argument.

    Args:
        policy_type: A key in `policy_registry`, e.g. `ppo`.
        venv: An environment that the policy is to be used with.
        **kwargs: Additional arguments to pass to the policy loader.

    Returns:
        The deserialized policy.
    """
    agent_loader = policy_registry.get(policy_type)
    return agent_loader(venv, **kwargs)


def save_stable_model(
    output_dir: pathlib.Path,
    model: base_class.BaseAlgorithm,
    filename: str = "model.zip",
) -> None:
    """Serialize Stable Baselines model.

    Load later with `load_policy(..., policy_path=output_dir)`.

    Args:
        output_dir: Path to the save directory.
        model: The stable baselines model.
        filename: The filename of the model.
    """
    # Save each model in new directory in case we want to add metadata or other
    # information in future. (E.g. we used to save `VecNormalize` statistics here,
    # although that is no longer necessary.)
    output_dir.mkdir(parents=True, exist_ok=True)
    model.save(output_dir / filename)
    logging.info(f"Saved policy to {output_dir}")


class SavePolicyCallback(callbacks.EventCallback):
    """Saves the policy using `save_stable_model` each time it is called.

    Should be used in conjunction with `callbacks.EveryNTimesteps`
    or another event-based trigger.
    """

    def __init__(
        self,
        policy_dir: pathlib.Path,
        *args,
        **kwargs,
    ):
        """Builds SavePolicyCallback.

        Args:
            policy_dir: Directory to save checkpoints.
            *args: Passed through to `callbacks.EventCallback`.
            **kwargs: Passed through to `callbacks.EventCallback`.
        """
        super().__init__(*args, **kwargs)
        self.policy_dir = policy_dir

    def _on_step(self) -> bool:
        assert self.model is not None
        output_dir = self.policy_dir / f"{self.num_timesteps:012d}"
        save_stable_model(output_dir, self.model)
        return True



"""Fixtures common across algorithm tests."""
from typing import Sequence

import pytest
from stable_baselines3.common import envs
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.vec_env import DummyVecEnv, VecEnv

from imitation.algorithms import bc
from imitation.data.types import TrajectoryWithRew
from imitation.data.wrappers import RolloutInfoWrapper
from imitation.policies import serialize
from imitation.testing.expert_trajectories import (
    lazy_generate_expert_trajectories,
    make_expert_transition_loader,
)
from imitation.util import util

CARTPOLE_ENV_NAME = "seals/CartPole-v0"


@pytest.fixture
def cartpole_expert_policy(cartpole_venv: VecEnv) -> BasePolicy:
    return serialize.load_policy(
        "ppo-huggingface",
        cartpole_venv,
        env_name=CARTPOLE_ENV_NAME,
    )


@pytest.fixture
def cartpole_expert_trajectories(
    cartpole_expert_policy,
    cartpole_venv,
    pytestconfig,
    rng,
) -> Sequence[TrajectoryWithRew]:
    return lazy_generate_expert_trajectories(
        pytestconfig.cache.makedir("experts"),
        CARTPOLE_ENV_NAME,
        60,
        rng,
    )


PENDULUM_ENV_NAME = "Pendulum-v1"


@pytest.fixture
def cartpole_bc_trainer(
    pytestconfig,
    cartpole_venv,
    cartpole_expert_trajectories,
    rng,
):
    return bc.BC(
        observation_space=cartpole_venv.observation_space,
        action_space=cartpole_venv.action_space,
        batch_size=50,
        demonstrations=make_expert_transition_loader(
            cache_dir=pytestconfig.cache.makedir("experts"),
            batch_size=50,
            expert_data_type="transitions",
            env_name="seals/CartPole-v0",
            rng=rng,
            num_trajectories=60,
        ),
        custom_logger=None,
        rng=rng,
    )


@pytest.fixture
def pendulum_expert_trajectories(
    pytestconfig,
    rng,
) -> Sequence[TrajectoryWithRew]:
    return lazy_generate_expert_trajectories(
        pytestconfig.cache.makedir("experts"),
        PENDULUM_ENV_NAME,
        60,
        rng,
    )


@pytest.fixture
def pendulum_expert_policy(pendulum_venv) -> BasePolicy:
    return serialize.load_policy(
        "ppo-huggingface",
        pendulum_venv,
        env_name=PENDULUM_ENV_NAME,
    )


@pytest.fixture
def pendulum_venv(rng) -> VecEnv:
    return util.make_vec_env(
        PENDULUM_ENV_NAME,
        n_envs=8,
        post_wrappers=[lambda env, _: RolloutInfoWrapper(env)],
        rng=rng,
    )


@pytest.fixture
def pendulum_single_venv(rng) -> VecEnv:
    return util.make_vec_env(
        PENDULUM_ENV_NAME,
        n_envs=1,
        post_wrappers=[lambda env, _: RolloutInfoWrapper(env)],
        rng=rng,
    )


@pytest.fixture
def multi_obs_venv() -> VecEnv:
    def make_env():
        env = envs.SimpleMultiObsEnv(channel_last=False)
        return RolloutInfoWrapper(env)

    return DummyVecEnv([make_env, make_env])



"""Configuration for imitation.scripts.train_adversarial."""

import pathlib

import sacred

from imitation.rewards import reward_nets
from imitation.scripts.ingredients import demonstrations, environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, reward, rl

# Note: All the hyperparameter configs in the file are of the tuned
# hyperparameters of the RL algorithm of the respective environment.
# Taken from imitation/scripts/config/train_rl.py

train_adversarial_ex = sacred.Experiment(
    "train_adversarial",
    ingredients=[
        logging_ingredient.logging_ingredient,
        demonstrations.demonstrations_ingredient,
        reward.reward_ingredient,
        rl.rl_ingredient,
        expert.expert_ingredient,
        environment.environment_ingredient,
        policy_evaluation.policy_evaluation_ingredient,
    ],
)


@train_adversarial_ex.config
def defaults():
    show_config = False

    total_timesteps = int(1e6)  # Num of environment transitions to sample
    algorithm_kwargs = dict(
        demo_batch_size=1024,  # Number of expert samples per discriminator update
        n_disc_updates_per_round=4,  # Num discriminator updates per generator round
    )
    algorithm_specific = {}  # algorithm_specific[algorithm] is merged with config

    checkpoint_interval = 0  # Num epochs between checkpoints (<0 disables)
    agent_path = None  # Path to load agent from, optional.


@train_adversarial_ex.config
def aliases_default_gen_batch_size(algorithm_kwargs, rl):
    # Setting generator buffer capacity and discriminator batch size to
    # the same number is equivalent to not using a replay buffer at all.
    # "Disabling" the replay buffer seems to improve convergence speed, but may
    # come at a cost of stability.
    algorithm_kwargs["gen_replay_buffer_capacity"] = rl["batch_size"]


# Shared settings

MUJOCO_SHARED_LOCALS = dict(rl=dict(rl_kwargs=dict(ent_coef=0.1)))

ANT_SHARED_LOCALS = dict(
    total_timesteps=int(3e7),
    algorithm_kwargs=dict(shared=dict(demo_batch_size=8192)),
    rl=dict(batch_size=16384),
)


# Classic RL Gym environment named configs


@train_adversarial_ex.named_config
def acrobot():
    environment = dict(gym_id="Acrobot-v1")
    algorithm_kwargs = {"allow_variable_horizon": True}


@train_adversarial_ex.named_config
def cartpole():
    environment = dict(gym_id="CartPole-v1")
    algorithm_kwargs = {"allow_variable_horizon": True}


@train_adversarial_ex.named_config
def seals_cartpole():
    environment = dict(gym_id="seals/CartPole-v0")
    total_timesteps = int(1.4e6)


@train_adversarial_ex.named_config
def mountain_car():
    environment = dict(gym_id="MountainCar-v0")
    algorithm_kwargs = {"allow_variable_horizon": True}


@train_adversarial_ex.named_config
def seals_mountain_car():
    environment = dict(gym_id="seals/MountainCar-v0")


@train_adversarial_ex.named_config
def pendulum():
    environment = dict(gym_id="Pendulum-v1")


# Standard MuJoCo Gym environment named configs


CHEETAH_SHARED_LOCALS = dict(
    MUJOCO_SHARED_LOCALS,
    rl=dict(batch_size=16384, rl_kwargs=dict(batch_size=1024)),
    algorithm_specific=dict(
        airl=dict(total_timesteps=int(5e6)),
        gail=dict(total_timesteps=int(8e6)),
    ),
    reward=dict(
        algorithm_specific=dict(
            airl=dict(
                net_cls=reward_nets.BasicShapedRewardNet,
                net_kwargs=dict(
                    reward_hid_sizes=(32,),
                    potential_hid_sizes=(32,),
                ),
            ),
        ),
    ),
    algorithm_kwargs=dict(
        # Number of discriminator updates after each round of generator updates
        n_disc_updates_per_round=16,
        # Equivalent to no replay buffer if batch size is the same
        gen_replay_buffer_capacity=16384,
        demo_batch_size=8192,
    ),
)


@train_adversarial_ex.named_config
def half_cheetah():
    locals().update(**CHEETAH_SHARED_LOCALS)
    environment = dict(gym_id="HalfCheetah-v2")


@train_adversarial_ex.named_config
def seals_humanoid():
    locals().update(**MUJOCO_SHARED_LOCALS)
    environment = dict(gym_id="seals/Humanoid-v0")
    total_timesteps = int(4e6)


@train_adversarial_ex.named_config
def reacher():
    environment = dict(gym_id="Reacher-v2")
    algorithm_kwargs = {"allow_variable_horizon": True}


# Debug configs


@train_adversarial_ex.named_config
def fast():
    # Minimize the amount of computation. Useful for test cases.

    # Need a minimum of 10 total_timesteps for adversarial training code to pass
    # "any update happened" assertion inside training loop.
    total_timesteps = 10
    algorithm_kwargs = dict(
        demo_batch_size=1,
        n_disc_updates_per_round=4,
    )


hyperparam_dir = pathlib.Path(__file__).absolute().parent / "tuned_hps"
tuned_alg_envs = [
    "airl_seals_ant",
    "airl_seals_half_cheetah",
    "airl_seals_hopper",
    "airl_seals_swimmer",
    "airl_seals_walker",
    "gail_seals_ant",
    "gail_seals_half_cheetah",
    "gail_seals_hopper",
    "gail_seals_swimmer",
    "gail_seals_walker",
]

for tuned_alg_env in tuned_alg_envs:
    config_file = hyperparam_dir / f"{tuned_alg_env}_best_hp_eval.json"
    assert config_file.is_file(), f"{config_file} does not exist"
    train_adversarial_ex.add_named_config(tuned_alg_env, str(config_file))



r"""Generate commands to run training scripts with different configs.

This program takes as input a set of config files and outputs a command
for each config file. Each command runs a training script using the
associated config file. By default, each command uses the same random seed
for training. However, if multiple random seeds are specified, the program
generates multiple commands for each config file (one for each random seed).
The commands can be executed by copying and pasting them into a command line
or by piping them to another command line utility. This program is helpful
for running a large number of training scripts with different config files
using multiple random seeds.

For example, we can run:

TUNED_HPS_DIR=../src/imitation/scripts/config/tuned_hps
python commands.py \
    --name=run0 \
    --cfg_pattern=$TUNED_HPS_DIR/*ai*_seals_walker_*.json \
    --output_dir=output

And get the following commands printed out:

python -m imitation.scripts.train_adversarial airl \
    --capture=sys --name=run0 \
    --file_storage=output/sacred/$USER-cmd-run0-airl-0-a3531726 \
    with ../src/imitation/scripts/config/tuned_hps/airl_seals_walker_best_hp_eval.json \
    seed=0 logging.log_root=output

python -m imitation.scripts.train_adversarial gail \
    --capture=sys --name=run0 \
    --file_storage=output/sacred/$USER-cmd-run0-gail-0-a1ec171b \
    with $TUNED_HPS_DIR/gail_seals_walker_best_hp_eval.json \
    seed=0 logging.log_root=output

We can execute commands in parallel by piping them to GNU parallel:

python commands.py ... | parallel -j 8

If the --remote flag is enabled, then the program prints out commands
to run training scripts in containers on the Hofvarpnir cluster.

For example, we can run:

TUNED_HPS_DIR=../src/imitation/scripts/config/tuned_hps
python commands.py \
    --name=run0 \
    --cfg_pattern=$TUNED_HPS_DIR/bc_seals_half_cheetah_best_hp_eval.json \
    --output_dir=/data/output \
    --remote

And get the following command printed out:

ctl job run --name $USER-cmd-run0-bc-0-72cb1df3 \
    --command "python -m imitation.scripts.train_imitation bc \
    --capture=sys --name=run0 \
    --file_storage=/data/output/sacred/$USER-cmd-run0-bc-0-72cb1df3 with \
    /data/imitation/src/imitation/scripts/config/tuned_hps/
    bc_seals_half_cheetah_best_hp_eval.json \
    seed=0 logging.log_root=/data/output" \
    --container hacobe/devbox:imitation \
    --login --force-pull --never-restart --gpu 0 --shared-host-dir-mount /data
"""
import argparse
import glob
import os
import zlib

_ALGO_NAME_TO_SCRIPT_NAME = {
    "bc": "train_imitation",
    "dagger": "train_imitation",
    "airl": "train_adversarial",
    "gail": "train_adversarial",
}

_CMD_ID_TEMPLATE = "$USER-cmd-{name}-{algo_name}-{seed}-{cfg_id}"

_TRAIN_CMD_TEMPLATE = """python -m imitation.scripts.{script_name} \
{algo_name} --capture=sys --name={name} --file_storage={file_storage} \
with {cfg_path} seed={seed} logging.log_root={log_root}"""

_HOFVARPNIR_CLUSTER_CMD_TEMPLATE = """ctl job run \
--name {name} --command "{command}" --container {container} \
--login --force-pull --never-restart --gpu 0 \
--shared-host-dir-mount /data"""


def _get_algo_name(cfg_file: str) -> str:
    """Get the algorithm name from the given config filename."""
    algo_names = set()
    for key in _ALGO_NAME_TO_SCRIPT_NAME:
        if cfg_file.find(key + "_") != -1:
            algo_names.add(key)

    if len(algo_names) == 0:
        raise ValueError("Unable to find algo_name in cfg_file: " + cfg_file)

    if len(algo_names) >= 2:
        raise ValueError("algo_name is ambiguous in cfg_file: " + cfg_file)

    algo_name = algo_names.pop()
    return algo_name


def _get_cfg_id(cfg_path: str) -> str:
    """Get an ID for the config from the given config path."""
    checksum = zlib.adler32(cfg_path.encode())
    checksum_hex = hex(checksum)
    assert checksum_hex.startswith("0x")
    return checksum_hex[2:]


def main(args: argparse.Namespace) -> None:
    """Generate commands to run training scripts with different configs."""
    cfg_relative_paths = glob.glob(args.cfg_pattern)
    local = not args.remote

    for cfg_relative_path in cfg_relative_paths:
        cfg_file = os.path.basename(cfg_relative_path)
        algo_name = _get_algo_name(cfg_file)
        script_name = _ALGO_NAME_TO_SCRIPT_NAME[algo_name]

        if local:
            cfg_path = cfg_relative_path
        else:
            cfg_path = os.path.join(args.remote_cfg_dir, cfg_file)

        cfg_id = _get_cfg_id(cfg_file)

        for seed in args.seeds:
            cmd_id = _CMD_ID_TEMPLATE.format(
                name=args.name,
                algo_name=algo_name,
                seed=seed,
                cfg_id=cfg_id,
            )

            file_storage = os.path.join(args.output_dir, "sacred", cmd_id)

            train_cmd = _TRAIN_CMD_TEMPLATE.format(
                script_name=script_name,
                algo_name=algo_name,
                name=args.name,
                cfg_path=cfg_path,
                seed=seed,
                file_storage=file_storage,
                log_root=args.output_dir,
            )

            if local:
                print(train_cmd)
                continue

            # Escape double quotes.
            command = train_cmd.replace('"', '\\"')

            hofvarpnir_cluster_cmd = _HOFVARPNIR_CLUSTER_CMD_TEMPLATE.format(
                name=cmd_id,
                command=command,
                container=args.container,
            )

            print(hofvarpnir_cluster_cmd)


def parse() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Generate commands to run training scripts with different configs.",
    )
    parser.add_argument(
        "--name",
        type=str,
        required=True,
        help="""A name that identifies multiple commands as \
coming from the same 'run'. In particular, this flag is passed to imitation training \
scripts directly in the --name flag and as part of the path in the \
--file_storage flag. If the --remote flag is enabled, this flag is also used in the \
cluster job name.""",
    )
    parser.add_argument(
        "--cfg_pattern",
        type=str,
        default="bc_seals_half_cheetah_best_hp_eval.json",
        help="""Generate a command for every file that matches this glob pattern. \
Each matching file should be a config file that has its algorithm name \
(bc, dagger, airl or gail) bookended by underscores in the filename. \
If the --remote flag is enabled, then generate a command for every file in the \
--remote_cfg_dir directory that has the same filename as a file that matches this \
glob pattern. E.g., suppose the current, local working directory is 'foo' and \
the subdirectory 'foo/bar' contains the config files 'bc_best.json' and \
'dagger_best.json'. If the pattern 'bar/*.json' is supplied, then globbing \
will return ['bar/bc_best.json', 'bar/dagger_best.json']. \
If the --remote flag is enabled, 'bar' will be replaced with `remote_cfg_dir` and \
commands will be created for the following configs: \
[`remote_cfg_dir`/bc_best.json, `remote_cfg_dir`/dagger_best.json] \
Why not just supply the pattern '`remote_cfg_dir`/*.json' directly? \
Because the `remote_cfg_dir` directory may not exist on the local machine.""",
    )
    parser.add_argument(
        "--seeds",
        type=int,
        nargs="+",
        default=[0],
        help="Space-delimited list of random seeds to use for each config.",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="""Directory in which to store the output training. \
If the --remote flag is enabled, this directory should be accessible \
from each container, e.g., '/data/output/' if /data is the shared directory.""",
    )
    parser.add_argument(
        "--remote",
        default=False,
        action="store_true",
        help="""Generate commands to run training scripts \
in containers on the Hofvarpnir cluster.""",
    )
    # The following flags are only used when the --remote flag is enabled.
    # Otherwise, they are ignored.
    parser.add_argument(
        "--remote_cfg_dir",
        type=str,
        default="/data/imitation/src/imitation/scripts/config/tuned_hps",
        help="""Path to a directory storing config files \
accessible from each container. """,
    )
    parser.add_argument(
        "--container",
        type=str,
        default="hacobe/devbox:imitation",
        help="The image name to use for the containers.",
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse()
    main(args)



"""Wrapper for reward labeling for transitions sampled from a replay buffer."""

from typing import Mapping, Type

import numpy as np
from gymnasium import spaces
from stable_baselines3.common.buffers import ReplayBuffer
from stable_baselines3.common.type_aliases import ReplayBufferSamples

from imitation.rewards.reward_function import RewardFn
from imitation.util import util


def _samples_to_reward_fn_input(
    samples: ReplayBufferSamples,
) -> Mapping[str, np.ndarray]:
    """Convert a sample from a replay buffer to a numpy array."""
    return dict(
        state=samples.observations.cpu().numpy(),
        action=samples.actions.cpu().numpy(),
        next_state=samples.next_observations.cpu().numpy(),
        done=samples.dones.cpu().numpy(),
    )


class ReplayBufferRewardWrapper(ReplayBuffer):
    """Relabel the rewards in transitions sampled from a ReplayBuffer."""

    def __init__(
        self,
        buffer_size: int,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        *,
        replay_buffer_class: Type[ReplayBuffer],
        reward_fn: RewardFn,
        **kwargs,
    ):
        """Builds ReplayBufferRewardWrapper.

        Args:
            buffer_size: Max number of elements in the buffer
            observation_space: Observation space
            action_space: Action space
            replay_buffer_class: Class of the replay buffer.
            reward_fn: Reward function for reward relabeling.
            **kwargs: keyword arguments for ReplayBuffer.
        """
        # Note(yawen-d): we directly inherit ReplayBuffer and leave out the case of
        # DictReplayBuffer because the current RewardFn only takes in NumPy array-based
        # inputs, and SAC is the only use case for ReplayBuffer relabeling. See:
        # https://github.com/HumanCompatibleAI/imitation/pull/459#issuecomment-1201997194
        assert replay_buffer_class is ReplayBuffer, "only ReplayBuffer is supported"
        assert not isinstance(observation_space, spaces.Dict)
        self.replay_buffer = replay_buffer_class(
            buffer_size,
            observation_space,
            action_space,
            **kwargs,
        )
        self.reward_fn = reward_fn
        _base_kwargs = {k: v for k, v in kwargs.items() if k in ["device", "n_envs"]}
        super().__init__(buffer_size, observation_space, action_space, **_base_kwargs)

    @property
    def pos(self) -> int:
        return self.replay_buffer.pos

    @pos.setter
    def pos(self, pos: int):
        self.replay_buffer.pos = pos

    @property
    def full(self) -> bool:
        return self.replay_buffer.full

    @full.setter
    def full(self, full: bool):
        self.replay_buffer.full = full

    def sample(self, *args, **kwargs):
        samples = self.replay_buffer.sample(*args, **kwargs)
        rewards = self.reward_fn(**_samples_to_reward_fn_input(samples))
        shape = samples.rewards.shape
        device = samples.rewards.device
        rewards_th = util.safe_to_tensor(rewards).reshape(shape).to(device)

        return ReplayBufferSamples(
            samples.observations,
            samples.actions,
            samples.next_observations,
            samples.dones,
            rewards_th,
        )

    def add(self, *args, **kwargs):
        self.replay_buffer.add(*args, **kwargs)

    def _get_samples(self):
        raise NotImplementedError(
            "_get_samples() is intentionally not implemented."
            "This method should not be called.",
        )



"""Tests examples/*: quickstart code and Jupyter notebook."""

import os
import pathlib
import subprocess
import sys
from typing import Iterable, Sequence

import pytest
import pytest_notebook as ptnb


def _paths_to_strs(x: Iterable[pathlib.Path]) -> Sequence[str]:
    """Convert Path to str for nice Pytest `parameterized` logs.

    For example, if we use Path, we get something inscrutable like
    test_run_example_sh_scripts[sh_path0] rather than seeing the actual path name.

    Args:
        x: The paths to convert.

    Returns:
        A sequence of the same length as `x`, with each element the string
        representation of the corresponding path in `x`.
    """
    return [str(path) for path in x]


THIS_DIR = pathlib.Path(__file__).absolute().parent
EXAMPLES_DIR = THIS_DIR / ".." / "examples"
TUTORIALS_DIR = THIS_DIR / ".." / "docs" / "tutorials"

EXCLUDED_EXAMPLE_FILES = ["train_dagger_atari_interactive_policy.py"]
EXCLUDED_EXAMPLE_PATHS = [EXAMPLES_DIR / f for f in EXCLUDED_EXAMPLE_FILES]

SH_PATHS = _paths_to_strs(set(EXAMPLES_DIR.glob("*.sh")) - set(EXCLUDED_EXAMPLE_PATHS))
TUTORIAL_PATHS = _paths_to_strs(TUTORIALS_DIR.glob("*.ipynb"))
PY_PATHS = _paths_to_strs(set(EXAMPLES_DIR.glob("*.py")) - set(EXCLUDED_EXAMPLE_PATHS))


# Note: This is excluded from coverage since is computed on linux. However, it is
#   covered by mac and windows runners.
@pytest.mark.skipif(sys.platform == "linux", reason="Linux is covered by readthedocs.")
@pytest.mark.parametrize("nb_path", TUTORIAL_PATHS)
def test_run_tutorial_notebooks(nb_path) -> None:  # pragma: no cover
    """Smoke test ensuring that tutorial notebooks run without error.

    The `pytest_notebook` package also includes regression test functionality against
    saved notebook outputs, if we want to check that later.

    Args:
        nb_path: Path to the notebook to test.
    """
    nb = ptnb.notebook.load_notebook(nb_path)
    # TODO(GH#793): Shorten timeout and ensure the notebook can still show desired
    # improvement.
    result = ptnb.execution.execute_notebook(nb, cwd=TUTORIALS_DIR, timeout=540)
    assert result.exec_error is None


@pytest.mark.parametrize("py_path", PY_PATHS)
def test_run_example_py_scripts(py_path):
    """Smoke test ensuring that python example scripts run without error."""
    # We need to use sys.executable, not just "python", on Windows as
    # subprocess.call ignores PATH (unless shell=True) so runs a
    # system-wide Python interpreter outside of our venv. See:
    # https://stackoverflow.com/questions/5658622/
    exit_code = subprocess.call([sys.executable, py_path])
    assert exit_code == 0


@pytest.mark.parametrize("sh_path", SH_PATHS)
def test_run_example_sh_scripts(sh_path):
    """Smoke test ensuring that shell example scripts run without error."""
    if os.name == "nt":  # pragma: no cover
        pytest.skip("bash shell scripts not ported to Windows.")
    for _ in range(2):  # Repeat because historically these have failed on second run.
        exit_code = subprocess.call(["env", "bash", "-e", sh_path])
        assert exit_code == 0


README_SNIPPET_PATHS = _paths_to_strs([EXAMPLES_DIR / "quickstart.sh"])


@pytest.mark.parametrize("snippet_path", README_SNIPPET_PATHS)
def test_example_snippets_are_in_readme(snippet_path):
    """Check that README.md examples haven't diverged from snippets."""
    with open(snippet_path, "r") as f:
        x = "".join(f.readlines()[2:])  # strip away shebang line
    with open("README.md", "r", encoding="utf-8") as f:
        y = f.read()
    assert x in y, f"{snippet_path} has diverged from README.md"
